# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import logging
import random

from variety.plugins.downloaders.DefaultDownloader import DefaultDownloader
from variety.Util import Util

logger = logging.getLogger("variety")


class ArtStationDownloader(DefaultDownloader):
    def __init__(self, source, url):
        DefaultDownloader.__init__(self, source=source, config=url)

    def fill_queue(self):
        logger.info(lambda: "ArtStation URL: " + self.config)

        queue = []
        # json_url = ArtStationDownloader.build_json_url(self.config)
        url = self.config
        s = Util.html_soup(url)
        author = s.find("channel").find("title").get_text().strip()
        author_url = s.find("channel").find("link").next.strip()
        items = s.findAll("item")
        for index, item in enumerate(items):
            try:
                extra_metadata = {
                    "headline": item.find("title").get_text().strip(),
                    "description": item.find("description").get_text().strip().replace("]]>", ""),
                    "author": author,
                    "authorURL": author_url,
                }
                src_url = item.find("guid").text + "#" + str(index)
                image_urls = [img["src"] for img in item.findAll("img")]
                for image_url in image_urls:
                    queue.append((src_url, image_url, extra_metadata))
            except Exception:
                logger.exception(lambda: "Could not process an item in the ArtStation rss result")

        random.shuffle(queue)
        return queue



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import random

from variety.plugins.downloaders.SimpleDownloader import SimpleDownloader
from variety.Util import Util, _

logger = logging.getLogger("variety")

random.seed()


class APODDownloader(SimpleDownloader):
    DESCRIPTION = _("NASA Astro Pic of the Day")
    ROOT_URL = "https://apod.nasa.gov/apod/"

    @classmethod
    def get_info(cls):
        return {
            "name": "APODDownloader",
            "description": APODDownloader.DESCRIPTION,
            "author": "Peter Levi",
            "version": "0.1",
        }

    def get_description(self):
        return APODDownloader.DESCRIPTION

    def get_source_type(self):
        return "apod"

    def get_source_name(self):
        return "NASA Astro Pic of the Day"

    def get_folder_name(self):
        return "nasa_apod"

    def get_source_location(self):
        return self.ROOT_URL

    def fill_queue(self):
        logger.info(lambda: "Filling APOD queue from Archive")

        s = Util.html_soup(self.ROOT_URL + "archivepix.html")
        urls = [
            self.ROOT_URL + x["href"]
            for x in s.findAll("a")
            if x["href"].startswith("ap") and x["href"].endswith(".html")
        ]
        urls = urls[:730]  # leave only last 2 years' pics
        urls = [url for url in urls if not self.is_in_banned(url)]

        queue = urls[:3]  # always put the latest 3 first
        urls = urls[3:]
        random.shuffle(urls)  # shuffle the rest
        queue.extend(urls)
        return queue

    def download_queue_item(self, queue_item):
        origin_url = queue_item
        logger.info(lambda: "APOD URL: " + origin_url)

        s = Util.html_soup(origin_url)
        img_url = None
        try:
            link = s.find("img").parent["href"]
            if link.startswith("image/"):
                img_url = self.ROOT_URL + link
                logger.info(lambda: "Image URL: " + img_url)
        except Exception:
            pass

        if img_url:
            return self.save_locally(origin_url, img_url, source_location=self.ROOT_URL)
        else:
            logger.info(lambda: "No image url found for this APOD URL")
            return None



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import abc

from variety.plugins.downloaders.ImageSource import ImageSource


class ConfigurableImageSource(ImageSource, metaclass=abc.ABCMeta):
    """
    Implements configurable/searchable image sources.
    """

    @abc.abstractmethod
    def validate(self, config):
        """
        Given a source config, validates whether images can be fetched for this config.
        As part of the validation, performs config normalization and initial lookup calls and
        returns a "final" config that will be persisted.

        Returns a tuple (config_to_persist, error_message)
        If error_message is None, the config is considered valid and a source is added.
        If error_message is not None, it is shown as an error in UI and a new source is not added.
        Example valid: return formatted(query), None
        Example invalid: return query, _('Not a proper XYZ query')
        @:param config the search query, URL or other config string for the new source
        @:return tuple (config_to_persist, error_message_or_none)
        """
        pass

    @abc.abstractmethod
    def create_downloader(self, config):
        """
        Creates a Downloader for the given config
        :param config: the downloader config
        :return: a Downloader object
        """
        pass

    @abc.abstractmethod
    def get_ui_instruction(self):
        """
        Long instruction to show in the "Add..." dialog
        :return: long instruction
        """
        pass

    @abc.abstractmethod
    def get_ui_short_instruction(self):
        """
        Short instructiton to show in the "Add..." dialog in front of the input field
        :return: short instruction
        """
        pass

    @abc.abstractmethod
    def get_ui_short_description(self):
        """
        Source description to show in the "Add..." menu popup
        :return: short description
        """
        pass



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import importlib.util
import sys
import inspect
import logging
import os

from .IPlugin import IPlugin

logger = logging.getLogger("variety")


class Jumble:
    def __init__(self, folders):
        self.folders = folders

    def _walk_python_files(self):
        for folder in self.folders:
            for location, directories, filenames in os.walk(folder):
                for f in filenames:
                    if f.endswith(".py"):
                        yield location, f

    def _walk_modules(self):
        for location, f in self._walk_python_files():
            path = os.path.join(location, f)
            name = os.path.splitext(f)[0]
            if (spec := importlib.util.spec_from_file_location(name, path)) is not None:
                try:
                    module = importlib.util.module_from_spec(spec)
                    logger.info(lambda: "Jumble loading module in %s from %s" % (name, path))
                    sys.modules[name] = module
                    spec.loader.exec_module(module)
                    yield module, path
                except Exception:
                    logger.exception("Could not load plugin module %s" % path)
                    continue

    def _walk_plugin_classes(self):
        for module, path in self._walk_modules():

            def is_plugin(cls):
                return (
                    inspect.isclass(cls)
                    and issubclass(cls, IPlugin)
                    and cls.__module__ == module.__name__
                )

            for name, cls in inspect.getmembers(module, is_plugin):
                yield cls, path

    def load(self):
        """
        Loads all plugins from the plugin folders, without activating them
        """
        logger.info(lambda: "Jumble loading")
        self.plugins = []
        for cls, path in self._walk_plugin_classes():
            try:
                info = cls.get_info()
            except Exception:
                logger.exception("Jumble: not a plugin class: %s" % str(cls))
                continue

            if not info:
                logger.warning("Jumble: %s: get_info() returned None" % str(cls))
                continue

            try:
                plugin = cls()
                logger.info(lambda: "Jumble found plugin class: %s: %s" % (str(cls), str(info)))

                plugin.jumble = self
                plugin.path = os.path.realpath(path)
                plugin.folder = os.path.dirname(plugin.path)
                self.plugins.append({"plugin": plugin, "class": cls, "info": info})
            except Exception:
                logger.exception("Jumble: could not instantiate plugin class: %s" % str(cls))
                continue

    def get_plugins(self, clazz=None, typename=None, name=None, active=None):
        """
        Searches for plugins that match the given criteria. If no criteria are given, all loaded plugins are returned.

        :param clazz: parent plugin class; optional
        :param typename: plugin type name; optional
        :param name: plugin name, should match exactly; optional
        :param active: specifies whether the plugin should be currently active, or inactive; optional,
        by default both are returned
        :return: all matching plugins, as hashes {"plugin": plugin, "class": plugin_class, "info": info}
        """
        return sorted(
            [
                p
                for p in self.plugins
                if (not clazz or issubclass(p["class"], clazz))
                and (not typename or p["class"].__name__ == typename)
                and (not name or p["info"]["name"] == name)
                and (active is None or p["plugin"].is_active() == active)
            ],
            key=lambda p: p["info"]["name"],
        )



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import threading

import cairo
from PIL import Image

from variety.Util import Util

# fmt: off
import gi  # isort:skip
gi.require_version("PangoCairo", "1.0")
from gi.repository import Gdk, GdkPixbuf, GObject, Pango, PangoCairo  # isort:skip
# fmt: on


class QuoteWriter:
    @staticmethod
    def write_quote(quote, author, infile, outfile, options=None):
        done_event = threading.Event()
        exception = [None]

        def go():
            try:
                w, h = Util.get_scaled_size(infile)
                surface = QuoteWriter.load_cairo_surface(infile, w, h)
                QuoteWriter.write_quote_on_surface(surface, quote, author, options)
                QuoteWriter.save_cairo_surface(surface, outfile)
            except Exception as e:
                exception[0] = e
            finally:
                done_event.set()

        Util.add_mainloop_task(go)
        done_event.wait()
        if exception[0]:
            raise exception[0]  # pylint: disable=raising-bad-type

    @staticmethod
    def load_cairo_surface(filename, w, h):
        # pylint: disable=no-member
        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(filename, w, h, False)
        surface = cairo.ImageSurface(0, pixbuf.get_width(), pixbuf.get_height())
        context = cairo.Context(surface)
        Gdk.cairo_set_source_pixbuf(context, pixbuf, 0, 0)
        context.paint()
        return surface

    @staticmethod
    def save_cairo_surface(surface, filename):
        try:
            # attempt faster method first
            # the get_data() call will fail with Cairo version < 1.15.4-1 (e.g. on 16.04)
            # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=809479
            data = surface.get_data()
            size = surface.get_width(), surface.get_height()
            image = Image.frombuffer("RGBA", size, data.tobytes(), "raw", "BGRA", 0, 1).convert(
                "RGB"
            )
            image.save(filename, quality=100)
        except:
            # fallback to slower method, but which works on 16.04
            surface.write_to_png(filename)

    @staticmethod
    def write_quote_on_surface(surface, quote, author=None, options=None, margin=30):
        qcontext = cairo.Context(surface)  # pylint: disable=no-member
        acontext = cairo.Context(surface)  # pylint: disable=no-member

        iw = surface.get_width()
        ih = surface.get_height()

        sw, sh = Util.get_primary_display_size(hidpi_scaled=True)
        trimw, trimh = Util.compute_trimmed_offsets((iw, ih), (sw, sh))

        width = max(
            200, sw * options.quotes_width // 100
        )  # use quotes_width percent of the visible width

        qlayout = PangoCairo.create_layout(qcontext)
        qlayout.set_width((width - 4 * margin) * Pango.SCALE)
        qlayout.set_alignment(Pango.Alignment.LEFT)
        qlayout.set_wrap(Pango.WrapMode.WORD)
        font = options.quotes_font if options else "Serif 30"
        qlayout.set_font_description(Pango.FontDescription(font))
        qlayout.set_text(quote, -1)

        qheight = qlayout.get_pixel_size()[1]
        qwidth = qlayout.get_pixel_size()[0]
        if options.quotes_width < 98:
            width = qwidth + 4 * margin
        else:
            width = sw

        alayout = PangoCairo.create_layout(acontext)
        aheight = 0
        if author:
            alayout.set_width(qwidth * Pango.SCALE)
            alayout.set_alignment(Pango.Alignment.RIGHT)
            alayout.set_wrap(Pango.WrapMode.WORD)
            alayout.set_font_description(Pango.FontDescription(font))
            alayout.set_text(author, -1)

            aheight = alayout.get_pixel_size()[1]

        height = qheight + aheight + 2.5 * margin

        bgc = options.quotes_bg_color
        qcontext.set_source_rgba(
            bgc[0] / 255.0, bgc[1] / 255.0, bgc[2] / 255.0, options.quotes_bg_opacity / 100.0
        )  # gray semi-transparent background

        hpos = trimw + (sw - width) * options.quotes_hpos // 100
        vpos = trimh + (sh - height) * options.quotes_vpos // 100
        qcontext.rectangle(hpos, vpos, width, height)
        qcontext.fill()

        qcontext.translate(hpos + (width - qwidth) / 2, vpos + margin)

        if options.quotes_text_shadow:
            qcontext.set_source_rgba(0, 0, 0, 0.2)
            PangoCairo.update_layout(qcontext, qlayout)
            PangoCairo.show_layout(qcontext, qlayout)
            qcontext.translate(-2, -2)

        tc = options.quotes_text_color

        qcontext.set_source_rgb(tc[0] / 255.0, tc[1] / 255.0, tc[2] / 255.0)
        PangoCairo.update_layout(qcontext, qlayout)
        PangoCairo.show_layout(qcontext, qlayout)

        acontext.translate(hpos + (width - qwidth) / 2, vpos + margin + qheight + margin / 2)

        if options.quotes_text_shadow:
            acontext.set_source_rgba(0, 0, 0, 0.2)
            PangoCairo.update_layout(acontext, alayout)
            PangoCairo.show_layout(acontext, alayout)
            acontext.translate(-2, -2)

        acontext.set_source_rgb(tc[0] / 255.0, tc[1] / 255.0, tc[2] / 255.0)
        PangoCairo.update_layout(acontext, alayout)
        PangoCairo.show_layout(acontext, alayout)

        qcontext.show_page()
        acontext.show_page()


if __name__ == "__main__":
    QuoteWriter.write_quote(
        '"I may be drunk, Miss, but in the morning I will be sober and you will still be ugly."',
        "Winston Churchill",
        "test.jpg",
        "test_result.png",
    )



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import random
import re

from httplib2 import iri2uri
from variety.plugins.IQuoteSource import IQuoteSource
from variety.Util import Util, _

logger = logging.getLogger("variety")

KEYWORDS = [
    "love",
    "humor",
    "inspirational",
    "life",
    "funny",
    "writing",
    "death",
    "romance",
    "truth",
    "poetry",
    "religion",
    "god",
    "philosophy",
    "wisdom",
    "books",
    "happiness",
    "humour",
    "art",
    "faith",
    "politics",
    "reading",
    "science",
    "relationships",
    "war",
    "inspiration",
    "friendship",
    "women",
    "music",
    "success",
    "hope",
    "spirituality",
    "freedom",
    "fear",
    "beauty",
    "christianity",
    "history",
    "inspirational-quotes",
    "time",
    "marriage",
    "fantasy",
    "dreams",
    "sex",
    "nature",
    "education",
    "life-lessons",
    "change",
    "literature",
    "children",
    "pain",
    "fiction",
    "money",
    "people",
    "knowledge",
    "motivational",
    "family",
    "humanity",
    "peace",
    "reality",
    "kindlehighlight",
    "self-help",
    "men",
    "courage",
    "society",
    "creativity",
    "power",
    "humorous",
    "future",
    "food",
    "heart",
    "quotes",
    "work",
    "words",
    "memory",
    "leadership",
    "passion",
    "spiritual",
    "soul",
    "loss",
    "grief",
    "language",
    "psychology",
    "friends",
    "paranormal-romance",
    "learning",
    "imagination",
    "world",
    "magic",
    "sadness",
    "feminism",
    "depression",
]


class GoodreadsSource(IQuoteSource):
    def __init__(self):
        super(IQuoteSource, self).__init__()

    @classmethod
    def get_info(cls):
        return {
            "name": "Goodreads",
            "description": _("Fetches quotes from Goodreads.com"),
            "author": "Peter Levi",
            "version": "0.1",
        }

    def supports_search(self):
        return True

    def get_random(self):
        return self.get_for_keyword(random.choice(KEYWORDS))[:4]

    def get_for_author(self, author):
        logger.info(lambda: "Fetching quotes from Goodreads for author=%s" % author)

        url = iri2uri("https://www.goodreads.com/quotes/search?utf8=\u2713&q=%s" % author)
        soup = Util.html_soup(url)
        page_links = list(
            Util.safe_map(
                int,
                [
                    pagelink.contents[0]
                    for pagelink in soup.find_all(href=re.compile("quotes/search.*page="))
                ],
            )
        )
        if page_links:
            page = random.randint(1, max(page_links))
            url = iri2uri(
                "https://www.goodreads.com/quotes/search?utf8=\u2713&q=%s&page=%d" % (author, page)
            )
            soup = Util.html_soup(url)

        return self.get_from_soup(url, soup)

    def get_for_keyword(self, keyword):
        logger.info(lambda: "Fetching quotes from Goodreads for keyword=%s" % keyword)

        url = iri2uri("https://www.goodreads.com/quotes/tag?utf8=\u2713&id=%s" % keyword)
        soup = Util.html_soup(url)
        page_links = list(
            Util.safe_map(
                int,
                [
                    pagelink.contents[0]
                    for pagelink in soup.find_all(href=re.compile("quotes/tag.*page="))
                ],
            )
        )
        if page_links:
            page = random.randint(1, max(page_links))
            url = iri2uri(
                "https://www.goodreads.com/quotes/tag?utf8=\u2713&id=%s&page=%d" % (keyword, page)
            )
            soup = Util.html_soup(url)

        return self.get_from_soup(url, soup)

    def get_from_soup(self, url, soup):
        logger.info(lambda: "Used Goodreads url %s" % url)
        quotes = []

        for div in soup.find_all("div", {"class": "quoteText"}):
            logger.debug(lambda: "Parsing quote for div\n%s" % div)
            try:
                quote_text = (
                    "\n".join(div.find_all(text=True, recursive=False)).replace("―", "").strip()
                )
                author = (
                    div.find("span", attrs={"class": "authorOrTitle"}).string.strip().strip(",")
                )
                first_a = div.find("a")
                if first_a:
                    link = "https://www.goodreads.com" + first_a["href"]
                else:
                    link = None  # No link given
                quotes.append(
                    {"quote": quote_text, "author": author, "sourceName": "Goodreads", "link": link}
                )
            except Exception:
                logger.exception(lambda: "Could not extract Goodreads quote")

        if not quotes:
            logger.warning(lambda: "Goodreads: no quotes found at %s" % url)

        return quotes



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import os.path

from gi.repository import Gtk

from variety.Util import Util, _

logger = logging.getLogger("variety")


class FolderChooser:
    def __init__(self, button, on_change=None):
        self.button = button
        self.on_change = on_change
        self.folder = None
        self.chooser = None
        self.box = Gtk.Box(Gtk.Orientation.HORIZONTAL)
        self.label = Gtk.Label()
        self.image = Gtk.Image()
        self.image.set_margin_left(2)
        self.image.set_margin_right(5)
        self.image.set_from_icon_name("folder", Gtk.IconSize.MENU)
        self.box.add(self.image)
        self.box.add(self.label)
        self.box.set_margin_right(2)
        if self.button.get_child():
            self.button.get_child().destroy()
        self.button.add(self.box)
        self.button.show_all()
        self.button.connect("clicked", self.browse)

    def destroy(self):
        if self.chooser:
            self.chooser.destroy()

    def get_folder(self):
        return self.folder

    def set_folder(self, folder):
        self.folder = os.path.normpath(folder)
        self.image.set_from_icon_name(Util.get_file_icon_name(self.folder), Gtk.IconSize.MENU)
        self.label.set_text(Util.collapseuser(self.folder))
        self.button.set_tooltip_text(self.folder)

    def set_sensitive(self, sensitive):
        self.button.set_sensitive(sensitive)

    def browse(self, widget=None):
        try:
            self.chooser = Gtk.FileChooserDialog(
                _("Choose a folder"),
                parent=self.button.get_toplevel(),
                action=Gtk.FileChooserAction.SELECT_FOLDER,
                buttons=[_("Cancel"), Gtk.ResponseType.CANCEL, _("OK"), Gtk.ResponseType.OK],
            )
            self.chooser.set_filename(self.folder)
            self.chooser.set_select_multiple(False)
            self.chooser.set_local_only(False)

            if self.chooser.run() == Gtk.ResponseType.OK:
                self.set_folder(self.chooser.get_filename())
                try:
                    if self.on_change:
                        self.on_change()
                except Exception:
                    logger.exception(lambda: "Exception during FolderChooser on_change:")
        finally:
            if self.chooser:
                try:
                    self.chooser.destroy()
                    self.chooser = None
                except Exception:
                    logger.exception(lambda: "Exception during FolderChooser destroying:")



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2019, Dan Jones
# Copyright (c) 2019, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

#
# Based on gist: https://gist.github.com/goodevilgenius/3878ce0f3e232e3daf5c
#

import re
import subprocess
from locale import gettext as _

from variety.plugins.IQuoteSource import IQuoteSource

ANSI_ESCAPE_RE = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]", flags=re.IGNORECASE)


class FortuneSource(IQuoteSource):
    @classmethod
    def get_info(cls):
        return {
            "name": "UNIX fortune program",
            "description": _(
                "Displays quotes using the UNIX fortune program. "
                "You may want to install additional fortune packs, e.g. fortunes-bofh-excuses."
            ),
            "author": "Dan Jones",
            "version": "0.1",
        }

    def needs_internet(self):
        return False

    def get_random(self):
        fortune = subprocess.check_output(["fortune"]).decode().strip()
        fortune = ANSI_ESCAPE_RE.sub("", fortune)
        q = fortune.split("--")
        quote = q[0].strip()
        author = q[1].strip() if len(q) > 1 else None
        return [
            {"quote": quote, "author": author, "sourceName": "UNIX fortune program", "link": None}
        ]



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import os
import threading
import webbrowser

from gi.repository import Gdk, GdkPixbuf, GObject, Gtk

from configobj import ConfigObj
from variety.ThumbsWindow import ThumbsWindow
from variety.Util import Util, _
from variety_lib import varietyconfig

logger = logging.getLogger("variety")


class ThumbsManager:
    POSITIONS = {
        "bottom": ThumbsWindow.BOTTOM,
        "top": ThumbsWindow.TOP,
        "left": ThumbsWindow.LEFT,
        "right": ThumbsWindow.RIGHT,
    }

    POSITION_NAMES = {
        "bottom": _("Bottom"),
        "top": _("Top"),
        "left": _("Left"),
        "right": _("Right"),
    }

    R_POSITIONS = {v: k for (k, v) in POSITIONS.items()}

    SIZES = [x * 30 for x in range(2, 11)]

    UNLIMITED = "Unlimited"

    LIMITS = [10, 50, 100, 200, 500, 1000, 2000, UNLIMITED]

    class Options:
        def __init__(self):
            self.position = ThumbsWindow.BOTTOM
            self.breadth = 120
            self.limit = 200

    def __init__(self, parent):
        self.parent = parent
        self.thumbs_window = None
        self.show_thumbs_lock = threading.Lock()

        self.pinned = False
        self.images = []
        self.screen = None

        self.type = None
        self.folders = None

        self.active_file = None
        self.active_position = None

    def create_menu(self, file):
        options = self.load_options()

        menu = Gtk.Menu()

        position_menu = Gtk.Menu()
        for p, v in ThumbsManager.POSITIONS.items():
            item = Gtk.CheckMenuItem(ThumbsManager.POSITION_NAMES[p])
            item.set_draw_as_radio(True)
            item.set_active(options.position == v)

            def _set_position(widget, pos=p):
                self.set_position(pos)

            item.connect("activate", _set_position)
            position_menu.append(item)

        size_menu = Gtk.Menu()
        for size in ThumbsManager.SIZES:
            item = Gtk.CheckMenuItem(str(size))
            item.set_draw_as_radio(True)
            item.set_active(options.breadth == size)

            def _set_size(widget, size=size):
                self.set_size(size)

            item.connect("activate", _set_size)
            size_menu.append(item)

        limit_menu = Gtk.Menu()
        for limit in ThumbsManager.LIMITS:
            item = Gtk.CheckMenuItem(str(limit))
            item.set_draw_as_radio(True)
            item.set_active(options.limit == limit)

            def _set_limit(widget, limit=limit):
                self.set_limit(limit)

            item.connect("activate", _set_limit)
            limit_menu.append(item)

        position_item = Gtk.MenuItem(_("Position"))
        position_item.set_submenu(position_menu)
        menu.append(position_item)

        size_item = Gtk.MenuItem(_("Size"))
        size_item.set_submenu(size_menu)
        menu.append(size_item)

        limit_item = Gtk.MenuItem(_("Maximum Shown Images"))
        limit_item.set_submenu(limit_menu)
        menu.append(limit_item)

        menu.append(Gtk.SeparatorMenuItem.new())

        open_file = Gtk.MenuItem(os.path.basename(file).replace("_", "__"))

        def _open_file(widget):
            self.parent.open_file(widget, file)

        open_file.connect("activate", _open_file)
        menu.append(open_file)

        open_folder = Gtk.MenuItem(_("Show Containing Folder"))

        def _open_folder(widget):
            self.parent.open_folder(widget, file)

        open_folder.connect("activate", _open_folder)
        menu.append(open_folder)

        info = Util.read_metadata(file)
        if info and "sourceURL" in info and "sourceName" in info:
            url = info["sourceURL"]
            source_name = info["sourceName"]
            if "Fetched" in source_name:
                label = _("Fetched: Show Origin")
            else:
                label = _("View at %s") % source_name

            if len(label) > 50:
                label = label[:50] + "..."

            show_origin = Gtk.MenuItem(label)

            def _show_origin(widget=None):
                logger.info(lambda: "Opening url: " + url)
                webbrowser.open_new_tab(url)

            show_origin.connect("activate", _show_origin)
            menu.append(show_origin)

        menu.append(Gtk.SeparatorMenuItem.new())
        rating_item = Gtk.MenuItem(_("Set EXIF Rating"))
        rating_item.set_submenu(ThumbsManager.create_rating_menu(file, self.parent))
        if not os.access(file, os.W_OK):
            rating_item.set_sensitive(False)
        menu.append(rating_item)

        menu.append(Gtk.SeparatorMenuItem.new())

        self.copy_to_favorites = Gtk.MenuItem(_("Copy to _Favorites"))
        self.copy_to_favorites.set_use_underline(True)

        def _copy_to_favorites(widget):
            self.parent.copy_to_favorites(widget, file)

        self.copy_to_favorites.connect("activate", _copy_to_favorites)
        menu.append(self.copy_to_favorites)

        self.move_to_favorites = Gtk.MenuItem(_("Move to _Favorites"))
        self.move_to_favorites.set_use_underline(True)

        def _move_to_favorites(widget):
            self.parent.move_to_favorites(widget, file)
            self.remove_image(file)

        self.move_to_favorites.connect("activate", _move_to_favorites)
        self.move_to_favorites.set_visible(False)
        menu.append(self.move_to_favorites)

        trash_item = Gtk.MenuItem(_("Delete to _Trash"))
        trash_item.set_use_underline(True)

        def _trash(widget):
            self.parent.move_to_trash(widget, file)

        trash_item.connect("activate", _trash)
        menu.append(trash_item)

        focus = Gtk.MenuItem(_("Where is it from?"))
        focus.set_sensitive(self.parent.get_source(file) is not None)

        def _focus(widget):
            self.parent.focus_in_preferences(widget, file)

        focus.connect("activate", _focus)
        menu.append(focus)

        menu.append(Gtk.SeparatorMenuItem.new())

        def close(widget):
            self.hide(force=True)

        close_item = Gtk.MenuItem(_("Close"))
        close_item.connect("activate", close)
        menu.append(close_item)

        menu.show_all()

        favs_op = self.parent.determine_favorites_operation(file)
        self.parent.update_favorites_menuitems(self, False, favs_op)

        return menu

    @staticmethod
    def create_rating_menu(file, main_window):
        def _set_rating_maker(rating):
            def _set_rating(widget, rating=rating):
                try:
                    Util.set_rating(file, rating)
                    main_window.on_rating_changed(file)
                except Exception:
                    logger.exception(lambda: "Could not set EXIF rating")
                    main_window.show_notification(_("Could not set EXIF rating"))

            return _set_rating

        try:
            actual_rating = Util.get_rating(file)
        except Exception:
            actual_rating = None

        rating_menu = Gtk.Menu()
        for rating in range(5, 0, -1):
            item = Gtk.CheckMenuItem("\u2605" * rating)
            item.set_draw_as_radio(True)
            item.set_active(actual_rating == rating)
            item.set_sensitive(not item.get_active())
            item.connect("activate", _set_rating_maker(rating))
            rating_menu.append(item)

        rating_menu.append(Gtk.SeparatorMenuItem.new())

        unrated_item = Gtk.CheckMenuItem(_("Unrated"))
        unrated_item.set_draw_as_radio(True)
        unrated_item.set_active(actual_rating is None or actual_rating == 0)
        unrated_item.set_sensitive(not unrated_item.get_active())
        unrated_item.connect("activate", _set_rating_maker(None))
        rating_menu.append(unrated_item)

        rejected_item = Gtk.CheckMenuItem(_("Rejected"))
        rejected_item.set_draw_as_radio(True)
        rejected_item.set_active(actual_rating is not None and actual_rating < 0)
        rejected_item.set_sensitive(not rejected_item.get_active())
        rejected_item.connect("activate", _set_rating_maker(-1))
        rating_menu.append(rejected_item)

        rating_menu.show_all()
        return rating_menu

    def repaint(self):
        self.hide(keep_settings=True)
        if self.images:
            self.show(
                self.unlimited_images, screen=self.screen, type=self.type, folders=self.folders
            )

    def set_position(self, position):
        logger.info(lambda: "Setting thumbs position " + str(position))
        options = self.load_options()
        options.position = ThumbsManager.POSITIONS[position]
        self.save_options(options)
        self.repaint()

    def set_size(self, size):
        logger.info(lambda: "Setting thumbs size " + str(size))
        options = self.load_options()
        options.breadth = size
        self.save_options(options)
        self.repaint()

    def set_limit(self, limit):
        logger.info(lambda: "Setting thumbs limit " + str(limit))
        options = self.load_options()
        options.limit = limit
        self.save_options(options)
        self.repaint()

    def pin(self, widget=None):
        self.pinned = True

    def on_click(self, thumbs_window, file, widget, event):
        file = file
        self.pin()

        def _resume_scrolling(menu=None):
            thumbs_window.resume_scrolling()

        thumbs_window.pause_scrolling()
        if event.button == 1:
            if self.is_showing("history"):
                index = [info["eventbox"] for info in thumbs_window.all].index(widget)
                self.parent.move_to_history_position(index)
            else:
                self.parent.set_wallpaper(file)
            _resume_scrolling()
        else:
            menu = self.create_menu(file)

            def _compute_position(*args, **kwargs):
                x, y = event.get_root_coords()[0], event.get_root_coords()[1]
                h = menu.get_preferred_height()[1]
                return x, y - h if y - h >= 40 else y, True

            menu.connect("deactivate", _resume_scrolling)
            menu.popup(None, None, _compute_position, None, 0, event.time)

    def mark_active(self, file=None, position=None):
        self.active_file = file
        self.active_position = position
        if self.thumbs_window:
            if self.is_showing("history"):
                self.thumbs_window.mark_active(position=position)
            else:
                self.thumbs_window.mark_active(file=file)

    def show(self, images, screen=None, type=None, folders=None):
        options = self.load_options()
        with self.show_thumbs_lock:
            self.unlimited_images = images
            self.type = type
            limit = len(self.unlimited_images) if options.limit == self.UNLIMITED else options.limit
            self.images = self.unlimited_images[:limit]
            self.screen = screen
            self.folders = folders

            def _go():
                try:
                    if self.thumbs_window:
                        self.thumbs_window.destroy()
                        self.thumbs_window = None

                    if len(self.images) > 0:
                        self.initialize_thumbs_window()
                except Exception:
                    logger.exception(lambda: "Could not create thumbs window:")

            Util.add_mainloop_task(_go)

    def initialize_thumbs_window(self):
        options = self.load_options()

        def _go():
            self.thumbs_window = ThumbsWindow(
                screen=self.screen, position=options.position, breadth=options.breadth
            )
            try:
                icon = varietyconfig.get_data_file("media", "variety.svg")
                self.thumbs_window.set_icon_from_file(icon)
            except Exception:
                logger.exception(lambda: "Could not set thumbs window icon")

            if self.type == "history":
                title = _("Variety History")
            elif self.type == "downloads":
                title = _("Variety Recent Downloads")
            else:
                title = _("Variety Images")

            self.thumbs_window.set_title(title)
            self.thumbs_window.connect("clicked", self.on_click)

            def _on_close(window, event):
                self.hide(force=True)

            self.thumbs_window.connect("delete-event", _on_close)

            self.mark_active(self.active_file, self.active_position)

            self.thumbs_window.start(self.images)

        Util.add_mainloop_task(_go)

    def load_options(self):
        options = ThumbsManager.Options()
        try:
            config = ConfigObj(os.path.join(self.parent.config_folder, "ui.conf"))
            try:
                s = config["thumbs_position"].lower()
                options.position = ThumbsManager.POSITIONS[s]
            except Exception:
                logger.exception(lambda: "Missing or bad thumbs_position option in ui.conf")

            try:
                options.breadth = int(config["thumbs_size"])
            except Exception:
                logger.exception(lambda: "Missing or bad thumbs_size option in ui.conf")

            try:
                options.limit = (
                    int(config["thumbs_limit"])
                    if config["thumbs_limit"] != self.UNLIMITED
                    else self.UNLIMITED
                )
            except Exception:
                logger.exception(lambda: "Missing or bad thumbs_limit option in ui.conf")
        except Exception:
            logger.exception(lambda: "Could not read ui.conf")

        self.save_options(options)
        return options

    def save_options(self, options):
        try:
            config = ConfigObj(os.path.join(self.parent.config_folder, "ui.conf"))
            try:
                config["thumbs_position"] = ThumbsManager.R_POSITIONS[options.position]
                config["thumbs_size"] = options.breadth
                config["thumbs_limit"] = str(options.limit)
                config.write()
            except Exception:
                logger.exception(lambda: "Missing or bad thumbs_position option in ui.conf")
        except Exception:
            logger.exception(lambda: "Could not save ui.conf")

    def hide(self, force=True, keep_settings=False):
        if force:
            self.pinned = False

        if self.pinned:
            return

        if not keep_settings:
            self.type = None
            self.images = []
            self.screen = None
            self.folders = None

        if self.thumbs_window:

            def _go():
                self.thumbs_window.destroy()
                self.thumbs_window = None

            Util.add_mainloop_task(_go)
            self.parent.update_indicator(auto_changed=False)

    def remove_image(self, file):
        self.images = [f for f in self.images if f != file]
        if self.thumbs_window:
            if self.thumbs_window.fits_in_screen(+1000):
                self.repaint()
            else:
                self.thumbs_window.remove_image(file)

    def add_image(self, file):
        self.images.insert(0, file)
        if not self.thumbs_window:
            self.initialize_thumbs_window()
        else:
            self.thumbs_window.add_image(file, at_front=True)

    def is_showing(self, type):
        return self.type == type

    def get_folders(self):
        return self.folders



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import logging
import random
from datetime import datetime
from urllib.parse import parse_qs, urlparse

from variety.plugins.downloaders.SimpleDownloader import SimpleDownloader
from variety.Util import Util, _

logger = logging.getLogger("variety")

random.seed()


class BingDownloader(SimpleDownloader):
    DESCRIPTION = _("Bing Photo of the Day")
    BING_JSON_URL = (
        "https://www.bing.com/HPImageArchive.aspx?format=js&idx=0&n=100&mkt=en-US"
    )  # n=100, but max 8 images are actually returned... Pity.

    @classmethod
    def get_info(cls):
        return {
            "name": "BingDownloader",
            "description": BingDownloader.DESCRIPTION,
            "author": "Peter Levi",
            "version": "0.1",
        }

    def get_source_type(self):
        return "bing"

    def get_description(self):
        return BingDownloader.DESCRIPTION

    def get_source_name(self):
        return "Bing"

    def get_source_location(self):
        return "https://www.bing.com/gallery/"

    def get_local_filename(self, url):
        return parse_qs(urlparse(url).query)["id"][0]

    def fill_queue(self):
        queue = []
        s = Util.fetch_json(BingDownloader.BING_JSON_URL)
        for item in s["images"]:
            try:
                if not item["wp"]:
                    # not marked as a wallpaper
                    continue

                image_url = "https://www.bing.com" + item["urlbase"] + "_UHD.jpg"

                copyrightlink = item["copyrightlink"]
                if not "javascript:" in copyrightlink:
                    src_url = copyrightlink
                else:
                    filename = item["url"].split("/")[-1]
                    name = filename[0 : filename.find("_EN")]
                    src_url = "https://www.bing.com/gallery/#images/%s" % name

                try:
                    date = datetime.strptime(item["startdate"], "%Y%m%d").strftime("%Y-%m-%d")
                except:
                    date = item["startdate"]
                extra_metadata = {
                    "sourceType": "bing",
                    "sfwRating": 100,
                    "headline": "Bing Photo of the Day, %s" % date,
                    "description": item["copyright"],
                }
                queue.append((src_url, image_url, extra_metadata))
            except:
                logger.exception(lambda: "Could not process an item in the Bing json result")

        random.shuffle(queue)
        return queue



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# Copyright (c) 2025, James Lu <james@overdrivenetworks.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

__all__ = ["get_data_file"]

__license__ = "GPL-3"
__version__ = "0.9.0"

import importlib.resources

def get_data_file(*path_segments):
    """Get the full path to a data file."""

    pkg_files_root = importlib.resources.files('variety') / 'data'
    return str(pkg_files_root.joinpath(*path_segments))

def get_version():
    return __version__



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import logging
import os
import urllib.parse

from PIL import Image
from requests.exceptions import HTTPError

from variety.Util import Util, _

logger = logging.getLogger("variety")


class ImageFetcher:
    @staticmethod
    def url_ok(url, use_whitelist, hosts_whitelist):
        try:
            p = urllib.parse.urlparse(url)
            if p.scheme in ["http", "https"]:
                if use_whitelist:
                    for host in hosts_whitelist:
                        h = host.strip().lower()
                        if h and p.netloc.lower().find(h) >= 0:
                            return True
                else:
                    return p.path.lower().endswith((".jpg", ".jpeg", ".png", ".tiff", ".avif", ".webp"))
                    # skip gif - they are usually small images
            return False
        except Exception:
            return False

    @staticmethod
    def fetch(
        url,
        to_folder,
        origin_url=None,
        source_type=None,
        source_location=None,
        source_name=None,
        extra_metadata=None,
        progress_reporter=lambda a, b: None,
        verbose=True,
    ):
        reported = verbose
        try:
            logger.info(lambda: "Trying to fetch URL %s to %s " % (url, to_folder))
            if verbose:
                progress_reporter(_("Fetching"), url)

            if url.startswith("javascript:"):
                if verbose:
                    progress_reporter(_("Not an image"), url)
                return None

            if url.find("://") < 0:
                url = "file://" + url

            r = Util.request(url, stream=True)
            if not "content-type" in r.headers:
                logger.info(lambda: "Unknown content-type for url " + url)
                if verbose:
                    progress_reporter(_("Not an image"), url)
                return None

            ct = r.headers["content-type"]
            if not ct.startswith("image/"):
                logger.info(lambda: "Unsupported content-type for url " + url + ": " + ct)
                if verbose:
                    progress_reporter(_("Not an image"), url)
                return None

            local_name = Util.get_local_name(r.url)
            if "content-disposition" in r.headers:
                cd = r.headers["content-disposition"]
                cd_name = ImageFetcher.extract_filename_from_content_disposition(cd)
                if cd_name:
                    local_name = cd_name

            filename = os.path.join(to_folder, local_name)
            if os.path.exists(filename):
                m = Util.read_metadata(filename)
                if m and m.get("imageURL") == url:
                    logger.info(lambda: "Local file already exists (%s)" % filename)
                    return filename
                else:
                    logger.info(
                        lambda: "File with same name already exists, but from different imageURL; renaming new download"
                    )
                    filename = Util.find_unique_name(filename)

            logger.info(lambda: "Fetching to " + filename)
            if not reported:
                reported = True
                progress_reporter(_("Fetching"), url)

            local_filepath_partial = filename + ".partial"
            with open(local_filepath_partial, "wb") as f:
                Util.request_write_to(r, f)

            try:
                img = Image.open(local_filepath_partial)
            except Exception:
                progress_reporter(_("Not an image"), url)
                Util.safe_unlink(local_filepath_partial)
                return None

            if img.size[0] < 400 or img.size[1] < 400:
                # too small - delete and do not use
                progress_reporter(_("Image too small, ignoring it"), url)
                Util.safe_unlink(local_filepath_partial)
                return None

            metadata = {
                "sourceType": source_type or "fetched",
                "sourceName": source_name or "Fetched",
                "sourceURL": origin_url or url,
                "imageURL": url,
            }
            if source_location:
                metadata["sourceLocation"] = source_location
            metadata.update(extra_metadata or {})
            Util.write_metadata(local_filepath_partial, metadata)

            os.rename(local_filepath_partial, filename)
            logger.info(lambda: "Fetched %s to %s." % (url, filename))
            return filename

        except Exception as e:
            # pylint: disable=no-member
            logger.exception(lambda: "Fetch failed for URL " + url)
            if reported:
                if isinstance(e, HTTPError) and e.response.status_code in (403, 404):
                    progress_reporter(
                        _("Sorry, got %s error...") % str(e.response.status_code),
                        _("This means the link is no longer valid"),
                    )
                else:
                    progress_reporter(
                        _("Fetch failed for some reason"),
                        _(
                            "To get more information, please run Variety from terminal with -v option and retry the action"
                        ),
                    )
            return None

    @staticmethod
    def extract_filename_from_content_disposition(cd):
        parts = cd.split(";")
        for p in parts:
            p = p.strip()
            if p.startswith("filename="):
                name = p[p.find("=") + 1 :]
                if name[0] in ['"', "'"]:
                    name = name[1:]
                if name[-1] in ['"', "'"]:
                    name = name[:-1]
                return name
        return None



import abc
import os

from jumble.IPlugin import IPlugin
from variety.Util import Util


class IVarietyPlugin(IPlugin, metaclass=abc.ABCMeta):
    """
    Variety-specific plugin interface
    """

    def activate(self):
        super(IVarietyPlugin, self).activate()
        self.config_folder = os.path.join(
            self.jumble.parent.config_folder, "pluginconfig/" + os.path.basename(self.folder)
        )
        Util.makedirs(self.config_folder)

    def get_config_folder(self):
        """
        :return: The config directory which the plugin can use to store config or cache files
        """
        return self.config_folder

    def on_variety_start_complete(self):
        """
        Called towards the end of VarietyWindow.start, when options are loaded and
        caches - created
        """
        pass



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import optparse

from variety.Util import _, safe_print
from variety_lib import varietyconfig


class VarietyOptionParser(optparse.OptionParser):
    """Override optparse.OptionParser to allow for errors in options without exiting"""

    def __init__(self, usage, version, report_errors=True):
        optparse.OptionParser.__init__(self, usage=usage, version=version)
        self.report_errors = report_errors

    def print_help(self, file=None):
        """print_help(file : file = stdout)

        Print an extended help message, listing all options and any
        help text provided with them, to 'file' (default stdout).
        """
        if file is None:
            safe_print(self.format_help())
        else:
            file.write(self.format_help().encode())

    def error(self, msg):
        if self.report_errors:
            optparse.OptionParser.error(self, msg)
        else:
            raise ValueError(msg)


def parse_options(arguments, report_errors=True):
    """Support for command line options"""
    usage = _(
        """%prog [options] [files or urls]

Passing local files will add them to Variety's queue.
Passing remote URLs will make Variety fetch them to Fetched folder and place them in the queue.

To set a specific wallpaper: %prog --set /some/local/image.jpg
"""
    )

    parser = VarietyOptionParser(
        usage=usage, version="%%prog %s" % varietyconfig.get_version(), report_errors=report_errors
    )

    parser.add_option(
        "--profile",
        action="store",
        dest="profile",
        help=_(
            "Profile name or full path to the configuration folder Variety should use. "
            "If not specified, this is ~/.config/variety/. "
            "If just a name is used instead of a full path, the profile folder will be "
            "~/.config/variety-profiles/<name>. "
            "Use only when initially starting Variety - changing the profile path requires "
            "restart. Several instances of Variety can be started when using different profiles, "
            "each with its own separate configuration. This can be used for example to control "
            "several different screens or workspaces under desktop environments like XFCE which "
            "allow this. To pass commands to a running instance, pass the same --profile "
            "argument as the one it was started with in subsequent commands."
        ),
        default=None,
    )

    parser.add_option(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        default=0,
        help=_(
            "Show logging messages (-vv to -vvvvv will profile various parts of Variety with increasing detail"
        ),
    )

    parser.add_option(
        "-q", "--quit", action="store_true", dest="quit", help=_("Make the running instance quit")
    )

    parser.add_option(
        "--get",
        "--get-wallpaper",
        "--current",
        "--show-current",
        action="store_true",
        dest="show_current",
        help=_(
            "Print the current wallpaper location. Used only when the application is already running."
        ),
    )

    parser.add_option(
        "--meta",
        action="store_true",
        dest="show_meta",
        help=_(
            "Print the current wallpaper metadata. Used only when the application is already running."
        ),
    )

    parser.add_option(
        "--set",
        "--set-wallpaper",
        action="store",
        dest="set_wallpaper",
        help=_("Set the given file as wallpaper, absolute path required"),
    )

    parser.add_option(
        "-n", "--next", action="store_true", dest="next", help=_("Show Next wallpaper")
    )

    parser.add_option(
        "-p", "--previous", action="store_true", dest="previous", help=_("Show Previous wallpaper")
    )

    parser.add_option(
        "--fast-forward",
        action="store_true",
        dest="fast_forward",
        help=_("Show Next wallpaper, skipping the forward history"),
    )

    parser.add_option(
        "-t",
        "--trash",
        action="store_true",
        dest="trash",
        help=_(
            "Move current wallpaper to Trash. Used only when the application is already running."
        ),
    )

    parser.add_option(
        "-f",
        "--favorite",
        action="store_true",
        dest="favorite",
        help=_(
            "Copy current wallpaper to Favorites. Used only when the application is already running."
        ),
    )

    parser.add_option(
        "--move-to-favorites",
        action="store_true",
        dest="movefavorite",
        help=_(
            "Move current wallpaper to Favorites. Used only when the application is already running."
        ),
    )

    parser.add_option(
        "--pause", action="store_true", dest="pause", help=_("Pause on current image")
    )

    parser.add_option(
        "--resume", action="store_true", dest="resume", help=_("Resume regular image changes")
    )

    parser.add_option(
        "--toggle-pause",
        action="store_true",
        dest="toggle_pause",
        help=_("Toggle Pause/Resume state"),
    )

    parser.add_option(
        "--toggle-no-effects",
        action="store_true",
        dest="toggle_no_effects",
        help=_('Toggle "Show Without Effects" for current image'),
    )

    parser.add_option(
        "--quotes-next", action="store_true", dest="quotes_next", help=_("Show Next quote")
    )

    parser.add_option(
        "--quotes-previous",
        action="store_true",
        dest="quotes_previous",
        help=_("Show Previous quote"),
    )

    parser.add_option(
        "--quotes-fast-forward",
        action="store_true",
        dest="quotes_fast_forward",
        help=_("Show Next quote, skipping the forward history"),
    )

    parser.add_option(
        "--quotes-toggle-pause",
        action="store_true",
        dest="quotes_toggle_pause",
        help=_("Toggle Quotes Pause/Resume state"),
    )

    parser.add_option(
        "--quotes-save-favorite",
        action="store_true",
        dest="quotes_save_favorite",
        help=_("Save the current quote to Favorites"),
    )

    parser.add_option(
        "--history", action="store_true", dest="history", help=_("Toggle History display")
    )

    parser.add_option(
        "--downloads",
        action="store_true",
        dest="downloads",
        help=_("Toggle Recent Downloads display"),
    )

    parser.add_option(
        "--preferences",
        "--show-preferences",
        action="store_true",
        dest="preferences",
        help=_("Show Preferences dialog"),
    )

    parser.add_option(
        "--selector",
        "--show-selector",
        action="store_true",
        dest="selector",
        help=_(
            "Show manual wallpaper selector - the thumbnail bar filled with images from the active image sources"
        ),
    )

    parser.add_option(
        "--set-option",
        action="append",
        dest="set_options",
        nargs=2,
        help=_(
            "Sets and applies an option. "
            "The option names are the same that are used in Variety's config file "
            "~/.config/variety/variety.conf. "
            "Multiple options can be set in a single command. "
            "Example: 'variety --set-option icon Dark --set-option clock_enabled True'. "
            "USE WITH CAUTION: You are changing the settings file directly in an unguarded way."
        ),
    )

    options, args = parser.parse_args(arguments)

    if report_errors:
        if (options.next or options.fast_forward) and options.previous:
            parser.error(_("options --next/--fast-forward and --previous are mutually exclusive"))

        if options.trash and options.favorite:
            parser.error(_("options --trash and --favorite are mutually exclusive"))

        if options.pause and options.resume:
            parser.error(_("options --pause and --resume are mutually exclusive"))

        if (options.quotes_next or options.quotes_fast_forward) and options.quotes_previous:
            parser.error(
                _(
                    "options --quotes-next/--quotes-fast-forward and --quotes-previous are mutually exclusive"
                )
            )

    return options, args



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012-2018, Peter Levi <peterlevi@peterlevi.com>
# Copyright (c) 2017-2018, James Lu <james@overdrivenetworks.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import dbus, dbus.service, dbus.glib
import logging
import os
import signal
import sys

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject  # pylint: disable=E0611


class SafeLogger(logging.Logger):
    """
    Fixes UnicodeDecodeErrors errors in logging calls:
    Accepts lambda as well string messages. Catches errors when evaluating the passed lambda.
    """

    def makeRecord(self, name, level, fn, lno, msg, *args, **kwargs):
        try:
            new_msg = msg if isinstance(msg, str) else msg()
        except:
            locale_info = "Unknown"
            try:
                locale_info = "Terminal encoding=%s, LANG=%s, LANGUAGE=%s" % (
                    sys.stdout.encoding,
                    os.getenv("LANG"),
                    os.getenv("LANGUAGE"),
                )
                logging.getLogger("variety").exception(
                    "Errors while logging. Locale info: %s" % locale_info
                )
                # TODO gather and log more info here
            except:
                pass
            new_msg = "Errors while logging. Locale info: %s" % locale_info

        return super().makeRecord(name, level, fn, lno, new_msg, *args, **kwargs)


logging.setLoggerClass(SafeLogger)


# these must be after the setLoggerClass call, as they obtain the variety logger
from variety import VarietyWindow, ThumbsManager, ThumbsWindow
from variety.profile import set_profile_path, get_profile_path, is_default_profile, get_profile_id
from variety.Util import Util, _, ModuleProfiler, safe_print


# # Change default encoding from ascii to UTF8 - works OK on Linux and prevents various UnicodeEncodeErrors/UnicodeDecodeErrors
# Still, generally considerd bad practice, may cause some deep hidden errors, as various Python stuff depends on it
# reload(sys)
# sys.setdefaultencoding('UTF8')


def _get_dbus_key():
    """
    DBus key for Variety.
    Variety uses a different key per profile, so several instances can run simultaneously if
    running with different profiles.
    Command any instance from the terminal by passing explicitly the same --profile options as it
    was started with.
    :return: the dbus key
    """
    if is_default_profile():
        return "com.peterlevi.Variety"
    else:
        return "com.peterlevi.Variety_{}".format(get_profile_id())


DBUS_PATH = "/com/peterlevi/Variety"


class VarietyService(dbus.service.Object):
    def __init__(self, variety_window):
        self.variety_window = variety_window
        bus_name = dbus.service.BusName(_get_dbus_key(), bus=dbus.SessionBus())
        dbus.service.Object.__init__(self, bus_name, DBUS_PATH)

    @dbus.service.method(dbus_interface=_get_dbus_key(), in_signature="as", out_signature="s")
    def process_command(self, arguments):
        result = self.variety_window.process_command(arguments, initial_run=False)
        return "" if result is None else result


VARIETY_WINDOW = None

terminate = False


def _sigint_handler(*args):
    global terminate
    terminate = True


def _check_quit():
    global terminate
    if not terminate:
        GObject.timeout_add(1000, _check_quit)
        return

    logging.getLogger("variety").info("Terminating signal received, quitting...")
    safe_print(
        _("Terminating signal received, quitting..."),
        "Terminating signal received, quitting...",
        file=sys.stderr,
    )

    global VARIETY_WINDOW
    if VARIETY_WINDOW:
        VARIETY_WINDOW.on_quit()
    Util.start_force_exit_thread(10)


def _set_up_logging(verbose):
    # add a handler to prevent basicConfig
    root = logging.getLogger()
    null_handler = logging.NullHandler()
    root.addHandler(null_handler)

    formatter = logging.Formatter("%(levelname)s: %(asctime)s: %(funcName)s() '%(message)s'")

    logger = logging.getLogger("variety")
    logger_sh = logging.StreamHandler()
    logger_sh.setFormatter(formatter)
    logger.addHandler(logger_sh)

    try:
        logger_file = logging.FileHandler(os.path.join(get_profile_path(), "variety.log"), "w")
        logger_file.setFormatter(formatter)
        logger.addHandler(logger_file)
    except Exception:
        logger.exception("Could not create file logger")

    lib_logger = logging.getLogger("variety_lib")
    lib_logger_sh = logging.StreamHandler()
    lib_logger_sh.setFormatter(formatter)
    lib_logger.addHandler(lib_logger_sh)

    logger.setLevel(logging.INFO)
    # Set the logging level to show debug messages.
    if verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif not verbose:
        # If we're not in verbose mode, only log these messages to file. This prevents
        # flooding syslog and/or ~/.xsession-errors depending on how variety was started:
        # (https://bugs.launchpad.net/variety/+bug/1685003)
        # XXX: We should /really/ make the internal debug logging use logging.debug,
        # this is really just a bandaid patch.
        logger_sh.setLevel(logging.WARNING)

    if verbose >= 3:
        lib_logger.setLevel(logging.DEBUG)


def main():
    if os.geteuid() == 0:
        print(
            'Variety is not supposed to run as root.\n'
            'You should NEVER run desktop apps as root, unless they are supposed to make '
            'system-global changes and you know very well what you are doing.\n'
            'Please run it with your normal user.\n'
            '\n'
            'If you are trying to run as root because Variety does not start at all with your normal '
            'user, you may be hitting a file permission issue or a bug.\n'
            'Here is what to do to troubleshoot:\n'
            '\n'
            '1. Open a terminal and run "variety -v" with your normal user.\n'
            'Look for exceptions and hints in the log for what the problem might be.\n'
            '\n'
            '2. You may try to rename ~/.config/variety to ~/.config/variety_bak and try again.\n'
            'This will have Variety start from a clean state.\n'
            'Your old config and images will remain in variety_bak\n'
            '\n'
            '3. If none of these help, open a bug in https://github.com/varietywalls/variety/issues '
            'and follow the instructions there.'
        )
        sys.exit(1)

    # Ctrl-C
    signal.signal(signal.SIGINT, _sigint_handler)
    signal.signal(signal.SIGTERM, _sigint_handler)
    signal.signal(signal.SIGQUIT, _sigint_handler)

    arguments = sys.argv[1:]

    # validate arguments
    from variety import VarietyOptionParser

    options, args = VarietyOptionParser.parse_options(arguments)
    set_profile_path(options.profile)
    Util.makedirs(get_profile_path())

    # ensure singleton per profile
    bus = dbus.SessionBus()
    dbus_key = _get_dbus_key()
    if bus.request_name(dbus_key) != dbus.bus.REQUEST_NAME_REPLY_PRIMARY_OWNER:
        if not arguments or (options.profile and len(arguments) <= 2):
            arguments = ["--preferences"]
        safe_print(
            _("Variety is already running. Sending the command to the running instance."),
            "Variety is already running. Sending the command to the running instance.",
            file=sys.stderr,
        )
        method = bus.get_object(dbus_key, DBUS_PATH).get_dbus_method("process_command")
        result = method(arguments)
        if result:
            safe_print(result)
        return

    # set up logging
    # set_up_logging must be called after the DBus checks, only by one running instance,
    # or the log file can be corrupted
    _set_up_logging(options.verbose)
    logging.getLogger("variety").info(lambda: "Using profile folder {}".format(get_profile_path()))

    if options.verbose >= 3:
        profiler = ModuleProfiler()
        if options.verbose >= 5:
            # The main variety package
            pkgname = os.path.dirname(__file__)
            profiler.log_path(pkgname)

            if options.verbose >= 6:
                # Track variety_lib
                profiler.log_path(pkgname + "_lib")
        else:
            # Cherry-picked log items carried over from variety 0.6.x
            profiler.log_class(VarietyWindow.VarietyWindow)

            if options.verbose >= 4:
                profiler.log_class(ThumbsManager.ThumbsManager)
                profiler.log_class(ThumbsWindow.ThumbsWindow)

        profiler.start()

    # Run the application.
    window = VarietyWindow.VarietyWindow()
    global VARIETY_WINDOW
    VARIETY_WINDOW = window
    service = VarietyService(window)

    bus.call_on_disconnection(window.on_quit)

    window.start(arguments)
    GObject.timeout_add(2000, _check_quit)
    Gtk.main()



import os

from variety.Util import Util

DEFAULT_PROFILE_PATH = "~/.config/variety/"

__profile_path = DEFAULT_PROFILE_PATH


def set_profile_path(profile_path):
    if not profile_path:
        profile_path = DEFAULT_PROFILE_PATH

    # if just a name is passed instead of a full path, put it under ~/.config/variety-profiles
    if not "/" in profile_path:
        profile_path = "~/.config/variety-profiles/{}".format(profile_path)

    # make sure profile path has a trailing slash
    if not profile_path.endswith("/"):
        profile_path += "/"

    global __profile_path
    __profile_path = profile_path


def get_profile_path(expanded=True):
    global __profile_path
    return os.path.expanduser(__profile_path) if expanded else __profile_path


def get_profile_short_name():
    return os.path.basename(get_profile_path()[:-1])


def get_profile_wm_class():
    return "Variety" + ("" if is_default_profile() else " (Profile: {})".format(get_profile_path()))


def is_default_profile():
    """
    Are we using the default profile or a custom profile?
    """
    return os.path.normpath(get_profile_path()) == os.path.normpath(
        os.path.expanduser(DEFAULT_PROFILE_PATH)
    )


def get_profile_id():
    """
    Returns a dbus-and-filename-friendly identificator of the profile path
    """
    return Util.md5(os.path.normpath(get_profile_path()))[:10]


def get_desktop_file_name():
    if is_default_profile():
        return "variety.desktop"
    else:
        return "variety-{}-{}.desktop".format(get_profile_short_name(), get_profile_id())


def get_autostart_file_path():
    return os.path.join(os.path.expanduser("~/.config/autostart"), get_desktop_file_name())



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

from variety.AbstractAddByQueryDialog import AbstractAddByQueryDialog
from variety_lib.helpers import get_builder


class AddConfigurableDialog(AbstractAddByQueryDialog):
    __gtype_name__ = "AddConfigurableDialog"

    def __new__(cls):
        builder = get_builder("AddConfigurableDialog")
        new_object = builder.get_object("add_configurable_dialog")
        new_object.finish_initializing(builder)
        return new_object

    def set_source(self, source):
        self.source = source
        self.set_title("Variety - add {} source".format(source.get_source_name()))
        self.ui.title.set_text(source.get_source_name())
        self.ui.instruction.set_markup(source.get_ui_instruction())
        self.ui.short_instruction.set_markup(source.get_ui_short_instruction())

    def validate(self, query):
        return self.source.validate(query)

    def commit(self, final_query):
        if len(final_query):
            self.parent.on_add_dialog_okay(
                self.source.get_source_type(), final_query, self.edited_row
            )



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import abc
import json
import os

from variety.Util import Util


class Downloader(abc.ABC):
    def __init__(self, source, config=None):
        """
        Create a downloader for an image source
        :param source: the image source this downloader is associated with
        :param config: optional, see get_config()
        """
        super().__init__()
        self.source = source
        self.config = config
        self.target_folder = None  # initialized post construction by update_download_folder
        self.state = None

    def get_variety(self):
        """
        :return: the VarietyWindow instance, if set in the source, else None
        """
        return self.source.get_variety()

    def update_download_folder(self, global_download_folder):
        """
        Called after initialization, once the VarietyWindow instance is filled in, sets
        the target folder according to the current options. Also called if downloads folder
        is changed in settings.
        :param global_download_folder the "global" download folder as set in Variety's preferences
        Our target_folder will be a subfolder to that one, named depending on what get_folder_name()
        returns.
        :return the target_folder for this particular downloader
        """
        filename = self.get_folder_name()
        if len(filename) + len(global_download_folder) > 160:
            filename = filename[: (150 - len(global_download_folder))] + Util.md5(filename)[:10]
        self.target_folder = os.path.join(global_download_folder, filename)
        self._load_state()
        return self.target_folder

    def _load_state(self):
        try:
            with open(os.path.join(self.target_folder, "state.json")) as f:
                self.state = json.load(f)
        except Exception:
            self.state = {}

    def save_state(self):
        """
        Persists the state as json inside the downloader's target folder.
        state is a dict that is used internally by Variety, but the downloaders can also use it
        keeping any sort of state is necessary for the downloader.
        """
        if self.target_folder is None:
            raise Exception("update_download_folder was not called before save_state")
        with open(os.path.join(self.target_folder, "state.json"), "w") as f:
            json.dump(self.state, f)

    def get_local_filename(self, url):
        """
        Returns the local file name under which to save a remote image (at URL url).
        By default this is Util.get_local_name(url), but subclasses can override.
        :param url: the URL of the image
        :return: the full local path, under the current target_folder
        """
        return Util.get_local_name(url)

    def _local_filepath(self, url=None, local_filename=None):
        """
        Returns the local file path where to save a remote image (at URL url).
        By default this is os.path.join(self.target_folder, self.get_local_filename(url)).
        Subclasses are discouraged from overriding this method, override get_local_filename()
        instead so that downloaded files are kept under the downloader's target_folder.
        :param url: the URL of the image
        :param local_filename: what local filename to use instead of extracting it from the URL.
        Pass None to use self.get_local_filename(url)
        :return: the full local path, under the current target_folder
        """
        if url is None and local_filename is None:
            raise ValueError("One of url or local_filename must be specified")
        if self.target_folder is None:
            raise Exception("update_download_folder was not called before downloading")
        filename = local_filename if local_filename else self.get_local_filename(url)
        return os.path.join(self.target_folder, filename)

    def get_source(self):
        """
        Returns the IImageSource associated with this downloader
        :return: IImageSource
        """
        return self.source

    def get_source_type(self):
        """
        :return: the image source's source type
        """
        return self.source.get_source_type()

    def get_source_name(self):
        """
        :return: the image source's source name
        """
        return self.source.get_source_name()

    def get_source_location(self):
        """
        What to save as sourceLocation in the images' metadata.
        By default this is this downloader's config string.
        Override in SimpleDownloaders to return a non-null string.
        :return: the source location to be saved in images
        """
        return self.get_config()

    def get_description(self):
        """
        User-friendly description of this downloader to show in UI (in the list of sources).
        E.g. "Downloads random images from site X", or "Images from Flickr user XXX"
        :return: description
        """
        return self.config

    def get_config(self):
        """
        Returns the config string used to create this downloader, e.g. a search keyword, or URL
        :return: the config string, will be None for non-configurable sources
        """
        return self.config

    def get_identifier(self):
        return self.target_folder

    def get_folder_name(self):
        """
        How should the folder be named where images from this source are downloaded to
        :return: folder name (just name, not full path)
        """
        if self.config:
            return self.get_source_type() + "_" + Util.convert_to_filename(self.config)
        else:
            return self.get_source_name()

    def get_refresh_interval_seconds(self):
        """
        Refresher downloaders can download one and the same image URL on a regular basis.
        Returning a positive integer here instructs Variety to initiate regular downloads and
        wallpaper changes when using this downloader.
        Images downloaded by refreshers MUST contain the string "--refreshable" in the name -
        this is the way Variety will know these images can be set as wallpaper again and again.
        :return: False by defualt, override to return True when implementing refreshers
        :return: None, or a positive number of seconds
        """
        return None

    def is_refresher(self):
        """
        Checks if this is a "refresher" plugin, i.e. refresh_interval_seconds is not None
        """
        return self.get_refresh_interval_seconds() is not None

    @abc.abstractmethod
    def download_one(self):
        """
        Downloads a single image. DefaultDownloader provides a reference implementation and most
        plugins should inherit DefaultDownloader instead of implementing this method.
        :return: the downloaded file path, or None
        """
        pass



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import threading

from gi.repository import Gdk, Gtk  # pylint: disable=E0611

from variety.Util import Util


class AbstractAddByQueryDialog(Gtk.Dialog):
    def validate(self, query):
        """
        Example valid: return formatted(query), None
        Example invalid: return query, _('Not a proper XYZ query')
        """
        raise NotImplementedError()

    def commit(self, final_query):
        """
        Performs the actions to really add a valid query or URL to Variety
        Example: self.parent.on_add_dialog_okay(Options.SourceType.XYZ, final_query, self.edited_row)
        """
        raise NotImplementedError()

    def finish_initializing(self, builder):
        """Called when we're finished initializing.

        finish_initalizing should be called after parsing the ui definition
        and creating a AbstractAddByQueryDialog object with it in order to
        finish initializing the start of the new AbstractAddByQueryDialog
        instance.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self)
        self.edited_row = None

    def set_edited_row(self, edited_row):
        self.edited_row = edited_row
        self.ui.query.set_text(self.edited_row[2])

    def on_btn_ok_clicked(self, widget, data=None):
        if not len(self.ui.query.get_text().strip()):
            self.destroy()
        else:
            threading.Timer(0, self.ok_thread).start()

    def on_btn_cancel_clicked(self, widget, data=None):
        self.destroy()

    def ok_thread(self):
        def _start_ui():
            self.ui.message.set_visible(True)
            self.ui.buttonbox.set_sensitive(False)
            self.ui.query.set_sensitive(False)
            self.ui.spinner.set_visible(True)
            self.ui.spinner.start()
            self.ui.error.set_label("")

        Util.add_mainloop_task(_start_ui)

        query = self.ui.query.get_text().strip()

        final_query, invalid_msg = self.validate(query)

        def _stop_ui():
            if invalid_msg:
                self.ui.buttonbox.set_sensitive(True)
                self.ui.error.set_label(invalid_msg)
                self.ui.spinner.stop()
                self.ui.query.set_sensitive(True)
                self.ui.message.set_visible(False)
                self.ui.spinner.set_visible(False)
                self.ui.query.grab_focus()
            else:
                self.commit(final_query)
                self.destroy()

        Util.add_mainloop_task(_stop_ui)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import random
import re

import bs4
from httplib2 import iri2uri
from variety.plugins.IQuoteSource import IQuoteSource
from variety.Util import Util, _

logger = logging.getLogger("variety")


class QuotationsPageSource(IQuoteSource):
    @classmethod
    def get_info(cls):
        return {
            "name": "TheQuotationsPage.com",
            "description": _("Fetches quotes from TheQuotationsPage.com"),
            "author": "Peter Levi",
            "version": "0.1",
        }

    def supports_search(self):
        return True

    def get_from_html(self, url, html):
        quotes = []
        bs = bs4.BeautifulSoup(html, "lxml")
        fixmap = {
            "\u0091": "\u2018",
            "\u0092": "\u2019",
            "\u0093": "\u201C",
            "\u0094": "\u201D",
            "\u0085": "...",
            "\u0097": "\u2014",
            "\u0096": "-",
        }
        for item in bs.findAll("dt", "quote"):
            quote = None
            try:
                quote = item.find("a").contents[0]
                for k, v in fixmap.items():
                    quote = quote.replace(k, v)
                quote = "\u201C%s\u201D" % quote
                link = "http://www.quotationspage.com" + item.find("a")["href"]
                try:
                    author = item.next_sibling.find("b").find("a").contents[0]
                except Exception:
                    try:
                        author = item.next_sibling.find("b").contents[0]
                    except Exception:
                        author = None

                quotes.append(
                    {
                        "quote": quote,
                        "author": author,
                        "sourceName": "TheQuotationsPage.com",
                        "link": link,
                    }
                )
            except Exception:
                logger.warning(lambda: "Could not get or parse quote: %s" % quote)

        if not quotes:
            logger.warning(lambda: "QuotationsPage: no quotes found at %s" % url)

        return quotes

    def get_random(self):
        return self.get_for_search_url("http://www.quotationspage.com/random.php")

    def get_for_author(self, author):
        return self.get_for_search_url(
            iri2uri("http://www.quotationspage.com/search.php?Search=&Author=%s" % author)
        )

    def get_for_keyword(self, keyword):
        return self.get_for_search_url(
            iri2uri("http://www.quotationspage.com/search.php?Search=%s&Author=" % keyword)
        )

    def get_for_search_url(self, url):
        logger.info(lambda: "Fetching quotes from Goodreads for search url=%s" % url)
        html = Util.fetch(url)
        try:
            page = random.randint(1, int(re.findall(r"Page 1 of (\d+)", html)[0]))
            url += "&page=%d" % page
            html = Util.fetch(url)
        except Exception:
            pass  # probably just one page

        logger.info(lambda: "Used QuotationsPage url %s" % url)

        r = r".*<dl>(.*)</dl>.*"
        if re.match(r, html, flags=re.M | re.S):
            html = re.sub(r, "<html><body>\\1</body></html>", html, flags=re.M | re.S)
            # without this BeautifulSoup gets confused by some scripts

        return self.get_from_html(url, html)


if __name__ == "__main__":
    q = QuotationsPageSource()
    print(q.get_for_author("einstein"))
    print(q.get_for_keyword("funny"))
    print(q.get_random())



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging

from variety_lib.AboutDialog import AboutDialog

logger = logging.getLogger("variety")


# See variety_lib.AboutDialog.py for more details about how this class works.
class AboutVarietyDialog(AboutDialog):
    __gtype_name__ = "AboutVarietyDialog"

    def finish_initializing(self, builder):  # pylint: disable=E1002
        """Set up the about dialog"""
        super(AboutVarietyDialog, self).finish_initializing(builder)

        # Code for other initialization actions should be added here.



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import urllib
from urllib.parse import urlencode

from variety.AddConfigurableDialog import AddConfigurableDialog
from variety_lib.helpers import get_builder


class AddWallhavenDialog(AddConfigurableDialog):
    __gtype_name__ = "AddWallhavenDialog"

    def __new__(cls, variety):
        builder = get_builder("AddWallhavenDialog")
        new_object = builder.get_object("add_wallhaven_dialog")
        new_object.finish_initializing(builder)
        new_object.options = variety.options
        new_object.ui.apikey.set_text((variety.options.wallhaven_api_key or "").strip())
        return new_object

    def validate(self, query):
        api_key = self.ui.apikey.get_text().strip()
        query, msg = self.source.validate(query, api_key=api_key)
        if not msg:
            # valid query - persist the API key for future use
            self.options.wallhaven_api_key = api_key
            self.options.write()
        return query, msg

    @staticmethod
    def parse_location(location):
        url_parts = urllib.parse.urlparse(location)
        return urllib.parse.parse_qs(url_parts.query)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import abc


class IPlugin(object, metaclass=abc.ABCMeta):
    """
    The most simple interface to be inherited when creating a plugin.
    """

    @classmethod
    @abc.abstractmethod
    def get_info(cls):
        """
        Returns the basic info about the plugin. Please make sure the name is unique among all Variety plugins
        Format:
        return {
           "name": "Sample name",
           "description": "Sample description",
           "version": "1.0",
           "author": "Author name", # optional
           "url": "Plugin homepage URL"  # optional
        }
        """
        pass

    def __init__(self):
        """
        All plugins must have a default constructor with no parameters.
        Remember to call super.
        """
        self.active = False

        # These will be filled in by Jumble.load() and available before the first activate() call
        self.jumble = None
        self.path = None  # Path to the plugin python file
        self.folder = (
            None
        )  # Folder where plugin is located (can be used for loading UI resources, etc.).
        # This folder may be read-only. A separate config folder convention should be used to store config files.

    def activate(self):
        """
        Called at plugin activation. Please do not allocate large portions of memory or resources before this is called.
        Remember to call super first.
        This method can be called multiple times within a session.
        It may be called when the plugin is already active - in this case it should simply return.
        """
        if self.active:
            return
        self.active = True

    def deactivate(self):
        """
        Called when the plugin is disabled. Please free used memory and resources here.
        Remember to call super first.
        This method can be called multiple times within a session.
        It may be called when the plugin is already inactive - in this case it should simply return.
        """
        self.active = False

    def is_active(self):
        return self.active



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import logging
import random

from requests import HTTPError

from variety.plugins.builtin.downloaders.UnsplashDownloader import UnsplashDownloader
from variety.plugins.downloaders.ConfigurableImageSource import ConfigurableImageSource
from variety.plugins.downloaders.DefaultDownloader import DefaultDownloader
from variety.plugins.downloaders.ImageSource import Throttling
from variety.Util import Util, _

logger = logging.getLogger("variety")

random.seed()


class UnsupportedConfig(Exception):
    pass


class UnsplashConfigurableSource(ConfigurableImageSource):
    class UnsplashConfigurableDownloader(UnsplashDownloader):
        def __init__(self, source, config):
            DefaultDownloader.__init__(self, source, config)
            self.set_variety(source.get_variety())

        def get_source_type(self):
            return self.source.get_source_type()

        def get_description(self):
            return self.config

        def get_folder_name(self):
            return super(DefaultDownloader, self).get_folder_name()

        def get_unsplash_api_url(self):
            self.config = self.config.replace("http://", "https://")
            params = {}
            if self.config.startswith("https://"):
                if self.config.startswith("https://unsplash.com/s/photos/"):
                    query = self.config.replace("https://unsplash.com/s/photos/", "").split("/")[0]
                    params["query"] = query
                elif self.config.startswith("https://unsplash.com/@"):
                    username = self.config.replace("https://unsplash.com/@", "").split("/")[0]
                    params["username"] = username
                elif self.config.startswith("https://unsplash.com/collections/"):
                    collections = self.config.replace(
                        "https://unsplash.com/collections/", ""
                    ).split("/")[0]
                    params["collections"] = collections
                elif self.config.startswith("https://unsplash.com/t/"):
                    collections = self.config.replace("https://unsplash.com/t/", "").split("/")[0]
                    params["topics"] = collections
                else:
                    raise UnsupportedConfig()
            else:
                params["query"] = self.config

            return (
                super().get_unsplash_api_url()
                + "&"
                + "&".join("{}={}".format(key, value) for key, value in params.items())
            )

    @classmethod
    def get_info(cls):
        return {
            "name": "UnsplashConfigurableSource",
            "description": _("Configurable source for fetching photos from Unsplash.com"),
            "author": "Peter Levi",
            "version": "0.1",
        }

    def get_source_type(self):
        return "unsplash-search"

    def validate(self, config):
        try:
            url = self.UnsplashConfigurableDownloader(self, config).get_unsplash_api_url()
            data = Util.fetch_json(url)
            valid = "errors" not in data
            return config, None if valid else _("No images found")
        except UnsupportedConfig:
            return config, _("We do not support this type of URL")
        except Exception as e:
            if isinstance(e, HTTPError) and e.response.status_code == 404:
                return config, _("No images found")
            else:
                return config, _("Oops, this didn't work. Is the remote service up?")

    def create_downloader(self, config):
        return self.UnsplashConfigurableDownloader(self, config)

    def get_ui_instruction(self):
        return _(
            "We use the <a href='https://unsplash.com'>Unsplash</a> API to fetch random images "
            "that match the given search term or URL. The Unsplash API is rate-limited, so "
            "Variety fetches images from Unsplash sources as a reduced rate.\n"
            "\n"
            "Please specify either a search keyword, or the URL of a search result, user, "
            "collection or topic.\n"
            "Example: <a href='https://unsplash.com/collections/3863203/desktop-wallpapers'>https://unsplash.com/collections/3863203/desktop-wallpapers</a>\n"
            "Example: <a href='https://unsplash.com/@pawel_czerwinski'>https://unsplash.com/@pawel_czerwinski</a>\n"
            "Example: <a href='https://unsplash.com/t/wallpapers'>https://unsplash.com/t/wallpapers</a>"
        )

    def get_ui_short_instruction(self):
        return _("Search keyword or URL: ")

    def get_ui_short_description(self):
        return _("Fetch images from Unsplash.com for a given criteria")

    def get_source_name(self):
        return "Unsplash"

    def get_server_options_key(self):
        return "unsplash_v2"

    def get_default_throttling(self):
        return Throttling(max_downloads_per_hour=10, max_queue_fills_per_hour=1)

    def on_image_set_as_wallpaper(self, img, meta):
        return UnsplashDownloader().on_image_set_as_wallpaper(img, meta)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

# This is the preferences dialog.

import logging
import os
import random
import stat
import subprocess
import threading

from gi.repository import Gdk, GdkPixbuf, GObject, Gtk  # pylint: disable=E0611

from variety import Texts
from variety.AddConfigurableDialog import AddConfigurableDialog
from variety.AddWallhavenDialog import AddWallhavenDialog
from variety.EditFavoriteOperationsDialog import EditFavoriteOperationsDialog
from variety.FolderChooser import FolderChooser
from variety.Options import Options
from variety.plugins.IQuoteSource import IQuoteSource
from variety.profile import (
    get_autostart_file_path,
    get_profile_path,
    get_profile_short_name,
    get_profile_wm_class,
    is_default_profile,
)
from variety.Util import Util, _, on_gtk
from variety_lib import varietyconfig
from variety_lib.PreferencesDialog import PreferencesDialog
from variety_lib.varietyconfig import get_data_file

random.seed()
logger = logging.getLogger("variety")

SLIDESHOW_PAGE_INDEX = 4
DONATE_PAGE_INDEX = 10


class PreferencesVarietyDialog(PreferencesDialog):
    __gtype_name__ = "PreferencesVarietyDialog"

    def finish_initializing(self, builder, parent):  # pylint: disable=E1002
        """Set up the preferences dialog"""
        super(PreferencesVarietyDialog, self).finish_initializing(builder, parent)

        # Bind each preference widget to gsettings
        #        widget = self.builder.get_object('example_entry')
        #        settings.bind("example", widget, "text", Gio.SettingsBindFlags.DEFAULT)

        if Gdk.Screen.get_default().get_height() < 750:
            self.ui.sources_scrolled_window.set_size_request(0, 0)
            self.ui.hosts_scrolled_window.set_size_request(0, 0)
            self.ui.tips_scrolled_window.set_size_request(0, 0)

        PreferencesVarietyDialog.add_image_preview(self.ui.icon_chooser, 64)
        self.loading = False

        self.fav_chooser = FolderChooser(
            self.ui.favorites_folder_chooser, self.on_favorites_changed
        )
        self.fetched_chooser = FolderChooser(
            self.ui.fetched_folder_chooser, self.on_fetched_changed
        )
        self.copyto_chooser = FolderChooser(self.ui.copyto_folder_chooser, self.on_copyto_changed)
        self.slideshow_custom_chooser = FolderChooser(
            self.ui.slideshow_custom_chooser, self.delayed_apply
        )

        if not Util.check_variety_slideshow_present():
            self.ui.notebook.remove_page(SLIDESHOW_PAGE_INDEX)

        profile_suffix = (
            "" if is_default_profile() else _(" (Profile: {})").format(get_profile_short_name())
        )
        self.set_title(_("Variety Preferences") + profile_suffix)
        self.set_wmclass(get_profile_wm_class(), get_profile_wm_class())

        self.reload()

    def fill_smart_profile_url(self, msg):
        if "%SMART_PROFILE_URL%" in msg:
            profile_url = self.parent.smart.get_profile_url()
            msg = msg.replace("%SMART_PROFILE_URL%", profile_url) if profile_url else ""
        return msg

    def update_status_message(self):
        msg = ""
        if self.parent.server_options:
            try:
                msg_dict = self.parent.server_options.get("status_message", {})
                ver = varietyconfig.get_version()
                if ver in msg_dict:
                    msg = msg_dict[ver].strip()
                elif "*" in msg_dict:
                    msg = msg_dict["*"].strip()

                msg = self.fill_smart_profile_url(msg)
            except Exception:
                logger.exception(lambda: "Could not parse status message")
                msg = ""

        self.set_status_message(msg)

    @on_gtk
    def set_status_message(self, msg):
        self.ui.status_message.set_visible(msg)
        self.ui.status_message.set_markup(msg)

    def reload(self):
        try:
            logger.info(lambda: "Reloading preferences dialog")

            self.loading = True

            self.options = Options()
            self.options.read()

            self.ui.autostart.set_active(os.path.isfile(get_autostart_file_path()))

            self.ui.change_enabled.set_active(self.options.change_enabled)
            self.set_change_interval(self.options.change_interval)
            self.ui.change_on_start.set_active(self.options.change_on_start)
            self.ui.internet_enabled.set_active(self.options.internet_enabled)

            self.fav_chooser.set_folder(os.path.expanduser(self.options.favorites_folder))
            self.ui.wallpaper_auto_rotate.set_active(self.options.wallpaper_auto_rotate)
            self.ui.wallpaper_display_mode.remove_all()
            for mode in self.parent.get_display_modes():
                self.ui.wallpaper_display_mode.append(mode.id, mode.title)
            self.ui.wallpaper_display_mode.set_active_id(self.options.wallpaper_display_mode)

            self.fetched_chooser.set_folder(os.path.expanduser(self.options.fetched_folder))
            self.ui.clipboard_enabled.set_active(self.options.clipboard_enabled)
            self.ui.clipboard_use_whitelist.set_active(self.options.clipboard_use_whitelist)
            self.ui.clipboard_hosts.get_buffer().set_text("\n".join(self.options.clipboard_hosts))

            if self.options.icon == "Light":
                self.ui.icon.set_active(0)
            elif self.options.icon == "Dark":
                self.ui.icon.set_active(1)
            elif self.options.icon == "1":
                self.ui.icon.set_active(2)
            elif self.options.icon == "2":
                self.ui.icon.set_active(3)
            elif self.options.icon == "3":
                self.ui.icon.set_active(4)
            elif self.options.icon == "4":
                self.ui.icon.set_active(5)
            elif self.options.icon == "Current":
                self.ui.icon.set_active(6)
            elif self.options.icon == "None":
                self.ui.icon.set_active(8)
            else:
                self.ui.icon.set_active(7)
                self.ui.icon_chooser.set_filename(self.options.icon)

            if self.options.favorites_operations == [["/", "Copy"]]:
                self.ui.favorites_operations.set_active(0)
            elif self.options.favorites_operations == [["/", "Move"]]:
                self.ui.favorites_operations.set_active(1)
            elif self.options.favorites_operations == [["/", "Both"]]:
                self.ui.favorites_operations.set_active(2)
            else:
                self.ui.favorites_operations.set_active(3)

            self.favorites_operations = self.options.favorites_operations

            self.ui.change_lock_screen_box.set_active(self.options.change_lock_screen)
            self.ui.copyto_enabled.set_active(self.options.copyto_enabled)
            self.copyto_chooser.set_folder(self.parent.get_actual_copyto_folder())

            self.ui.desired_color_enabled.set_active(self.options.desired_color_enabled)
            self.ui.desired_color.set_color(
                Gdk.Color(red=160 * 256, green=160 * 256, blue=160 * 256)
            )
            c = self.options.desired_color
            if c:
                self.ui.desired_color.set_color(
                    Gdk.Color(red=c[0] * 256, green=c[1] * 256, blue=c[2] * 256)
                )

            self.ui.min_size_enabled.set_active(self.options.min_size_enabled)
            min_sizes = [50, 80, 100]
            index = 0
            while min_sizes[index] < self.options.min_size and index < len(min_sizes) - 1:
                index += 1
            self.ui.min_size.set_active(index)
            self.ui.landscape_enabled.set_active(self.options.use_landscape_enabled)
            self.ui.lightness_enabled.set_active(self.options.lightness_enabled)
            self.ui.lightness.set_active(
                0 if self.options.lightness_mode == Options.LightnessMode.DARK else 1
            )
            self.ui.min_rating_enabled.set_active(self.options.min_rating_enabled)
            self.ui.min_rating.set_active(self.options.min_rating - 1)
            self.ui.name_regex_enabled.set_active(self.options.name_regex_enabled)
            self.ui.name_regex.set_text(self.options.name_regex)
            self.ui.clock_enabled.set_active(self.options.clock_enabled)
            self.ui.clock_font.set_font_name(self.options.clock_font)
            self.ui.clock_date_font.set_font_name(self.options.clock_date_font)

            self.ui.quotes_enabled.set_active(self.options.quotes_enabled)
            self.ui.quotes_font.set_font_name(self.options.quotes_font)
            c = self.options.quotes_text_color
            self.ui.quotes_text_color.set_color(
                Gdk.Color(red=c[0] * 256, green=c[1] * 256, blue=c[2] * 256)
            )
            c = self.options.quotes_bg_color
            self.ui.quotes_bg_color.set_color(
                Gdk.Color(red=c[0] * 256, green=c[1] * 256, blue=c[2] * 256)
            )
            self.ui.quotes_bg_opacity.set_value(self.options.quotes_bg_opacity)
            self.ui.quotes_text_shadow.set_active(self.options.quotes_text_shadow)
            self.ui.quotes_tags.set_text(self.options.quotes_tags)
            self.ui.quotes_authors.set_text(self.options.quotes_authors)
            self.ui.quotes_change_enabled.set_active(self.options.quotes_change_enabled)
            self.set_quotes_change_interval(self.options.quotes_change_interval)
            self.ui.quotes_width.set_value(self.options.quotes_width)
            self.ui.quotes_hpos.set_value(self.options.quotes_hpos)
            self.ui.quotes_vpos.set_value(self.options.quotes_vpos)

            self.ui.slideshow_sources_enabled.set_active(self.options.slideshow_sources_enabled)
            self.ui.slideshow_favorites_enabled.set_active(self.options.slideshow_favorites_enabled)
            self.ui.slideshow_downloads_enabled.set_active(self.options.slideshow_downloads_enabled)
            self.ui.slideshow_custom_enabled.set_active(self.options.slideshow_custom_enabled)
            self.slideshow_custom_chooser.set_folder(
                os.path.expanduser(self.options.slideshow_custom_folder)
            )

            if self.options.slideshow_sort_order == "Random":
                self.ui.slideshow_sort_order.set_active(0)
            elif self.options.slideshow_sort_order == "Name, asc":
                self.ui.slideshow_sort_order.set_active(1)
            elif self.options.slideshow_sort_order == "Name, desc":
                self.ui.slideshow_sort_order.set_active(2)
            elif self.options.slideshow_sort_order == "Date, asc":
                self.ui.slideshow_sort_order.set_active(3)
            elif self.options.slideshow_sort_order == "Date, desc":
                self.ui.slideshow_sort_order.set_active(4)
            else:
                self.ui.slideshow_sort_order.set_active(0)

            self.ui.slideshow_monitor.remove_all()
            self.ui.slideshow_monitor.append_text(_("All"))

            screen = Gdk.Screen.get_default()
            for i in range(0, screen.get_n_monitors()):
                geo = screen.get_monitor_geometry(i)
                self.ui.slideshow_monitor.append_text(
                    "%d - %s, %dx%d"
                    % (i + 1, screen.get_monitor_plug_name(i), geo.width, geo.height)
                )
            self.ui.slideshow_monitor.set_active(0)
            try:
                self.ui.slideshow_monitor.set_active(int(self.options.slideshow_monitor))
            except:
                self.ui.slideshow_monitor.set_active(0)

            if self.options.slideshow_mode == "Fullscreen":
                self.ui.slideshow_mode.set_active(0)
            elif self.options.slideshow_mode == "Desktop":
                self.ui.slideshow_mode.set_active(1)
            elif self.options.slideshow_mode == "Maximized":
                self.ui.slideshow_mode.set_active(2)
            elif self.options.slideshow_mode == "Window":
                self.ui.slideshow_mode.set_active(3)
            else:
                self.ui.slideshow_mode.set_active(0)

            self.ui.slideshow_seconds.set_value(self.options.slideshow_seconds)
            self.ui.slideshow_fade.set_value(self.options.slideshow_fade)
            self.ui.slideshow_zoom.set_value(self.options.slideshow_zoom)
            self.ui.slideshow_pan.set_value(self.options.slideshow_pan)

            self.unsupported_sources = []
            self.ui.sources.get_model().clear()
            for s in self.options.sources:
                if s[1] in Options.get_all_supported_source_types():
                    self.ui.sources.get_model().append(self.source_to_model_row(s))
                else:
                    self.unsupported_sources.append(s)

            if not hasattr(self, "enabled_toggled_handler_id"):
                self.enabled_toggled_handler_id = self.ui.sources_enabled_checkbox_renderer.connect(
                    "toggled", self.source_enabled_toggled, self.ui.sources.get_model()
                )
            # self.ui.sources.get_selection().connect("changed", self.on_sources_selection_changed)

            if hasattr(self, "filter_checkboxes"):
                for cb in self.filter_checkboxes:  # pylint: disable=access-member-before-definition
                    self.ui.filters_grid.remove(cb)
                    cb.destroy()
            self.filter_checkboxes = []
            self.filter_name_to_checkbox = {}
            for i, f in enumerate(self.options.filters):
                cb = Gtk.CheckButton(Texts.FILTERS.get(f[1], f[1]))
                self.filter_name_to_checkbox[f[1]] = cb
                cb.connect("toggled", self.delayed_apply)
                cb.set_visible(True)
                cb.set_active(f[0])
                cb.set_margin_right(20)
                self.ui.filters_grid.attach(cb, i % 4, i // 4, 1, 1)
                self.filter_checkboxes.append(cb)

            # pylint: disable=access-member-before-definition
            if hasattr(self, "quotes_sources_checkboxes"):
                for cb in self.quotes_sources_checkboxes:
                    self.ui.quotes_sources_grid.remove(cb)
                    cb.destroy()
            self.quotes_sources_checkboxes = []
            for i, p in enumerate(self.parent.jumble.get_plugins(IQuoteSource)):
                cb = Gtk.CheckButton(p["info"]["name"])
                cb.connect("toggled", self.delayed_apply)
                cb.set_visible(True)
                cb.set_tooltip_text(p["info"]["description"])
                cb.set_active(p["info"]["name"] not in self.options.quotes_disabled_sources)
                cb.set_margin_right(20)
                self.ui.quotes_sources_grid.attach(cb, i % 4, i // 4, 1, 1)
                self.quotes_sources_checkboxes.append(cb)

            self.ui.tips_buffer.set_text(
                "\n\n".join(
                    [
                        tip.replace("{PROFILE_PATH}", get_profile_path(expanded=False))
                        for tip in Texts.TIPS
                    ]
                )
            )

            try:
                with open(get_data_file("ui/changes.txt")) as f:
                    self.ui.changes_buffer.set_text(f.read())
            except Exception:
                logger.warning(lambda: "Missing ui/changes.txt file")

            self.on_change_enabled_toggled()
            self.on_sources_selection_changed()
            self.on_desired_color_enabled_toggled()
            self.on_min_size_enabled_toggled()
            self.on_lightness_enabled_toggled()
            self.on_min_rating_enabled_toggled()
            self.on_name_regex_enabled_toggled()
            self.on_copyto_enabled_toggled()
            self.on_quotes_change_enabled_toggled()
            self.on_icon_changed()
            self.on_favorites_operations_changed()
            self.on_wallpaper_display_mode_changed()
            self.update_clipboard_state()
            self.update_status_message()
        finally:
            # To be sure we are completely loaded, pass via two hops: first delay, then idle_add:
            def _finish_loading():
                self.loading = False

            def _idle_finish_loading():
                Util.add_mainloop_task(_finish_loading)

            timer = threading.Timer(1, _idle_finish_loading)
            timer.start()

    def on_add_button_clicked(self, widget=None):
        def position(*args, **kwargs):
            button_alloc = self.ui.add_button.get_allocation()
            window_pos = self.ui.add_button.get_window().get_position()
            return (
                button_alloc.x + window_pos[0],
                button_alloc.y + button_alloc.height + window_pos[1],
                True,
            )

        add_menu = self.build_add_button_menu()
        add_menu.popup(None, self.ui.add_button, position, None, 0, Gtk.get_current_event_time())

    def on_remove_sources_clicked(self, widget=None):
        def position(*args, **kwargs):
            button_alloc = self.ui.remove_sources.get_allocation()
            window_pos = self.ui.remove_sources.get_window().get_position()
            return (
                button_alloc.x + window_pos[0],
                button_alloc.y + button_alloc.height + window_pos[1],
                True,
            )

        self.build_remove_button_menu().popup(
            None, self.ui.remove_sources, position, None, 0, Gtk.get_current_event_time()
        )

    def build_add_button_menu(self):
        add_menu = Gtk.Menu()

        items = [
            (False, _("Images"), _("Add individual wallpaper images"), self.on_add_images_clicked),
            (
                False,
                _("Folders"),
                _("Searched recursively for up to 10000 images, shown in random order"),
                lambda widget: self.on_add_folders_clicked(
                    widget, source_type=Options.SourceType.FOLDER
                ),
            ),
            (
                False,
                _("Sequential Albums (order by filename)"),
                _("Searched recursively for images, shown in sequence (by filename)"),
                lambda widget: self.on_add_folders_clicked(
                    widget, source_type=Options.SourceType.ALBUM_FILENAME
                ),
            ),
            (
                False,
                _("Sequential Albums (order by date)"),
                _("Searched recursively for images, shown in sequence (by file date)"),
                lambda widget: self.on_add_folders_clicked(
                    widget, source_type=Options.SourceType.ALBUM_DATE
                ),
            ),
            "-",
        ]

        configurable_items = []
        for source in self.options.CONFIGURABLE_IMAGE_SOURCES:

            def _click(widget, source=source):
                self.on_add_configurable(source)

            configurable_items.append(
                (
                    source.needs_internet(),
                    source.get_source_name(),
                    source.get_ui_short_description(),
                    _click,
                )
            )
        configurable_items.sort(key=lambda x: x[1])
        items.extend(configurable_items)

        for x in items:
            if x == "-":
                item = Gtk.SeparatorMenuItem.new()
                item.set_margin_top(15)
                item.set_margin_bottom(15)
            else:
                item = Gtk.MenuItem()
                label = Gtk.Label("<b>{}</b>\n{}".format(x[1], x[2]))
                label.set_margin_top(6)
                label.set_margin_bottom(6)
                label.set_xalign(0)
                label.set_use_markup(True)
                item.add(label)
                if x[0] and not self.ui.internet_enabled.get_active():
                    # disable adding internet-requiring sources when internet is disabled
                    item.set_sensitive(False)
                else:
                    item.connect("activate", x[3])
            add_menu.append(item)

        add_menu.show_all()
        return add_menu

    def build_remove_button_menu(self):
        model, rows = self.ui.sources.get_selection().get_selected_rows()

        has_downloaders = False
        for row in rows:
            type = model[row][1]
            if type in Options.get_editable_source_types():
                has_downloaders = True

        self.remove_menu = Gtk.Menu()
        item1 = Gtk.MenuItem()
        item1.set_label(
            _("Remove the source, keep the files")
            if len(rows) == 1
            else _("Remove the sources, keep the files")
        )
        item1.connect("activate", self.remove_sources)
        self.remove_menu.append(item1)

        item2 = Gtk.MenuItem()

        def _remove_with_files(widget=None):
            self.remove_sources(delete_files=True)

        item2.set_label(
            _("Remove the source and delete the downloaded files")
            if len(rows) == 1
            else _("Remove the sources and delete the downloaded files")
        )
        item2.connect("activate", _remove_with_files)
        item2.set_sensitive(has_downloaders)
        self.remove_menu.append(item2)

        self.remove_menu.show_all()
        return self.remove_menu

    def source_enabled_toggled(self, widget, path, model):
        row = model[path]
        row[0] = not row[0]
        self.on_row_enabled_state_changed(row)

    def on_row_enabled_state_changed(self, row):
        # Special case when enabling refresher downloaders:
        refresher_dls = [
            dl
            for dl in Options.SIMPLE_DOWNLOADERS  # TODO: this will break if we have non-simple refresher downloaders
            if dl.get_source_type() == row[1] and dl.is_refresher()
        ]
        if row[0] and len(refresher_dls) > 0:
            refresh_time = refresher_dls[0].get_refresh_interval_seconds()
            updated = False
            if not self.ui.change_enabled.get_active():
                self.ui.change_enabled.set_active(True)
                updated = True
            if self.get_change_interval() > refresh_time:
                self.set_change_interval(refresh_time)
                updated = True

            if updated:
                self.parent.show_notification(
                    refresher_dls[0].get_description(),
                    _(
                        "Using this source requires wallpaper changing "
                        "enabled at intervals of %d minutes or less. "
                        "Settings were adjusted automatically."
                    )
                    % int(refresh_time / 60),
                )

    def set_time(self, interval, text, time_unit, times=(1, 60, 60 * 60, 24 * 60 * 60)):
        if interval < 5:
            interval = 5
        x = len(times) - 1
        while times[x] > interval:
            x -= 1
        text.set_text(str(interval // times[x]))
        time_unit.set_active(x)
        return

    def set_change_interval(self, seconds):
        self.set_time(seconds, self.ui.change_interval_text, self.ui.change_interval_time_unit)

    def set_quotes_change_interval(self, seconds):
        self.set_time(
            seconds, self.ui.quotes_change_interval_text, self.ui.quotes_change_interval_time_unit
        )

    def read_time(self, text_entry, time_unit_combo, minimum, default):
        result = default
        try:
            interval = int(text_entry.get_text())
            tree_iter = time_unit_combo.get_active_iter()
            if tree_iter:
                model = time_unit_combo.get_model()
                time_unit_seconds = model[tree_iter][1]
                result = interval * time_unit_seconds
                if result < minimum:
                    result = minimum
        except Exception:
            logger.exception(lambda: "Could not understand interval")
        return result

    def get_change_interval(self):
        return self.read_time(
            self.ui.change_interval_text,
            self.ui.change_interval_time_unit,
            5,
            self.options.change_interval,
        )

    def get_quotes_change_interval(self):
        return self.read_time(
            self.ui.quotes_change_interval_text,
            self.ui.quotes_change_interval_time_unit,
            10,
            self.options.quotes_change_interval,
        )

    @staticmethod
    def add_image_preview(chooser, size=250):
        preview = Gtk.Image()
        chooser.set_preview_widget(preview)

        def update_preview(c):
            try:
                file = chooser.get_preview_filename()
                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(file, size, size)
                preview.set_from_pixbuf(pixbuf)
                chooser.set_preview_widget_active(True)
            except Exception:
                chooser.set_preview_widget_active(False)

        chooser.connect("update-preview", update_preview)

    def on_add_images_clicked(self, widget=None):
        chooser = Gtk.FileChooserDialog(
            _("Add Images"),
            parent=self,
            action=Gtk.FileChooserAction.OPEN,
            buttons=[_("Cancel"), Gtk.ResponseType.CANCEL, _("Add"), Gtk.ResponseType.OK],
        )
        self.dialog = chooser
        PreferencesVarietyDialog.add_image_preview(chooser)
        chooser.set_select_multiple(True)
        chooser.set_local_only(True)
        filter = Gtk.FileFilter()
        filter.set_name(_("Images"))
        for s in ["jpg", "jpeg", "png", "bmp", "tiff", "svg"]:
            filter.add_pattern("*." + s)
            filter.add_pattern("*." + s.upper())
        chooser.add_filter(filter)
        response = chooser.run()

        if response == Gtk.ResponseType.OK:
            images = list(chooser.get_filenames())
            images = [f for f in images if Util.is_image(f) and os.path.isfile(f)]
            self.add_sources(Options.SourceType.IMAGE, images)

        self.dialog = None
        chooser.destroy()

    def on_add_folders_clicked(self, widget=None, source_type=Options.SourceType.FOLDER):
        if source_type == Options.SourceType.FOLDER:
            title = _(
                "Add Folders - Only add the root folders, subfolders are searched recursively"
            )
        elif source_type == Options.SourceType.ALBUM_FILENAME:
            title = _(
                "Add Sequential Albums (ordered by filename). Subfolders are searched recursively."
            )
        elif source_type == Options.SourceType.ALBUM_DATE:
            title = _(
                "Add Sequential Albums (ordered by date). Subfolders are searched recursively."
            )
        else:
            raise Exception("Unsuppoted source_type {}".format(source_type))
        chooser = Gtk.FileChooserDialog(
            title,
            parent=self,
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            buttons=[_("Cancel"), Gtk.ResponseType.CANCEL, _("Add"), Gtk.ResponseType.OK],
        )
        self.dialog = chooser
        chooser.set_select_multiple(True)
        chooser.set_local_only(True)
        response = chooser.run()

        if response == Gtk.ResponseType.OK:
            folders = list(chooser.get_filenames())
            folders = [f for f in folders if os.path.isdir(f)]
            self.add_sources(source_type, folders)

        self.dialog = None
        chooser.destroy()

    def add_sources(self, type, locations):
        self.ui.sources.get_selection().unselect_all()
        existing = {}
        for i, r in enumerate(self.ui.sources.get_model()):
            if r[1] == type:
                if type in (
                    Options.SourceType.FOLDER,
                    Options.SourceType.ALBUM_FILENAME,
                    Options.SourceType.ALBUM_DATE,
                ):
                    existing[os.path.normpath(r[2])] = r, i
                else:
                    existing[self.model_row_to_source(r)[2]] = r, i

        newly_added = 0
        for f in locations:
            if type in Options.SourceType.LOCAL_PATH_TYPES:
                f = os.path.normpath(f)
            elif type not in Options.get_editable_source_types():
                f = (
                    list(existing.keys())[0] if existing else None
                )  # reuse the already existing location, do not add another one

            if not f in existing:
                self.ui.sources.get_model().append(self.source_to_model_row([True, type, f]))
                self.ui.sources.get_selection().select_path(len(self.ui.sources.get_model()) - 1)
                self.ui.sources.scroll_to_cell(
                    len(self.ui.sources.get_model()) - 1, None, False, 0, 0
                )
                newly_added += 1
            else:
                logger.info(lambda: "Source already exists, activating it: " + f)
                existing[f][0][0] = True
                self.ui.sources.get_selection().select_path(existing[f][1])
                self.ui.sources.scroll_to_cell(existing[f][1], None, False, 0, 0)

        return newly_added

    def focus_source_and_image(self, source, image):
        self.ui.notebook.set_current_page(0)
        self.ui.sources.get_selection().unselect_all()
        for i, r in enumerate(self.ui.sources.get_model()):
            if self.model_row_to_source(r)[1:] == source[1:]:
                self.focused_image = image
                self.ui.sources.get_selection().select_path(i)
                self.ui.sources.scroll_to_cell(i, None, False, 0, 0)
                return

    def remove_sources(self, widget=None, delete_files=False):
        model, rows = self.ui.sources.get_selection().get_selected_rows()

        if delete_files:
            for row in rows:
                type = model[row][1]
                if type in Options.get_editable_source_types():
                    source = self.model_row_to_source(model[row])
                    self.parent.delete_files_of_source(source)

        # store the treeiters from paths
        iters = []
        for row in rows:
            if model[row][1] in Options.get_removable_source_types():
                iters.append(model.get_iter(row))
        # remove the rows (treeiters)
        for i in iters:
            if i is not None:
                model.remove(i)

    def on_source_doubleclicked(self, tree_view, row_index, arg4=None):
        self.edit_source(self.ui.sources.get_model()[row_index])

    def on_edit_source_clicked(self, widget=None):
        model, rows = self.ui.sources.get_selection().get_selected_rows()
        if len(rows) == 1:
            self.edit_source(model[rows[0]])

    def on_open_folder_clicked(self, widget=None):
        model, rows = self.ui.sources.get_selection().get_selected_rows()
        if len(rows) != 1:
            return
        row = model[rows[0]]
        type = row[1]
        if type in Options.SourceType.LOCAL_PATH_TYPES:
            subprocess.Popen(["xdg-open", os.path.realpath(row[2])])
        elif type == Options.SourceType.FAVORITES:
            subprocess.Popen(["xdg-open", self.parent.options.favorites_folder])
        elif type == Options.SourceType.FETCHED:
            subprocess.Popen(["xdg-open", self.parent.options.fetched_folder])
        else:
            subprocess.Popen(
                ["xdg-open", self.parent.get_folder_of_source(self.model_row_to_source(row))]
            )

    def on_use_clicked(self, widget=None):
        model, rows = self.ui.sources.get_selection().get_selected_rows()
        for row in model:
            row[0] = False
        for path in rows:
            model[path][0] = True
        for row in model:
            # TODO we trigger for all rows, though some of them don't actually change state - but no problem for now
            self.on_row_enabled_state_changed(row)
        self.on_sources_selection_changed()

    def edit_source(self, edited_row):
        type = edited_row[1]

        if type in Options.get_editable_source_types():
            if type in Options.CONFIGURABLE_IMAGE_SOURCES_MAP:
                if type == Options.SourceType.WALLHAVEN:
                    self.dialog = AddWallhavenDialog(self.parent)
                else:
                    self.dialog = AddConfigurableDialog()
                self.dialog.set_source(Options.CONFIGURABLE_IMAGE_SOURCES_MAP[type])

            self.dialog.set_edited_row(edited_row)
            self.show_dialog(self.dialog)

    def on_internet_enabled_toggled(self, *args):
        self.delayed_apply()
        self.previous_selection = None
        self.on_sources_selection_changed()

    def on_sources_selection_changed(self, widget=None):
        model, rows = self.ui.sources.get_selection().get_selected_rows()

        enabled = set(i for i, row in enumerate(model) if row[0])
        selected = set(row.get_indices()[0] for row in rows)
        self.ui.use_button.set_sensitive(selected and enabled != selected)

        # pylint: disable=access-member-before-definition
        if hasattr(self, "previous_selection") and rows == self.previous_selection:
            return

        self.previous_selection = rows

        self.ui.edit_source.set_sensitive(False)
        self.ui.edit_source.set_label(_("Edit..."))
        self.ui.open_folder.set_sensitive(len(rows) == 1)
        self.ui.open_folder.set_label(_("Open Folder"))

        if len(rows) == 1:
            source = model[rows[0]]
            type = source[1]
            if type == Options.SourceType.IMAGE:
                self.ui.open_folder.set_label(_("View Image"))
            elif type in Options.get_editable_source_types():
                self.ui.edit_source.set_sensitive(self.ui.internet_enabled.get_active())

        def timer_func():
            self.show_thumbs(list(model[row] for row in rows))

        # pylint: disable=access-member-before-definition
        if hasattr(self, "show_timer") and self.show_timer:
            self.show_timer.cancel()
        self.show_timer = threading.Timer(0.3, timer_func)
        self.show_timer.start()

        for row in rows:
            if model[row][1] not in Options.get_removable_source_types():
                self.ui.remove_sources.set_sensitive(False)
                return

        self.ui.remove_sources.set_sensitive(len(rows) > 0)

    def model_row_to_source(self, row):
        return [row[0], row[1], Texts.SOURCES[row[1]][0] if row[1] in Texts.SOURCES else row[2]]

    def source_to_model_row(self, s):
        srctype = s[1]
        return [s[0], srctype, s[2] if not srctype in Texts.SOURCES else Texts.SOURCES[srctype][1]]

    def show_thumbs(self, source_rows, pin=False, thumbs_type=None):
        try:
            if not source_rows:
                return

            self.parent.thumbs_manager.hide(force=True)

            images = []
            folders = []

            for row in source_rows:
                if not row:
                    continue

                type = row[1]
                if type == Options.SourceType.IMAGE:
                    images.append(row[2])
                else:
                    folder = self.parent.get_folder_of_source(self.model_row_to_source(row))
                    folders.append(folder)

            folder_images = list(
                Util.list_files(folders=folders, filter_func=Util.is_image, max_files=10000)
            )
            if len(source_rows) == 1 and source_rows[0][1] == Options.SourceType.ALBUM_FILENAME:
                folder_images = sorted(folder_images)
            elif len(source_rows) == 1 and source_rows[0][1] == Options.SourceType.ALBUM_DATE:
                folder_images = sorted(folder_images, key=os.path.getmtime)
            else:
                random.shuffle(folder_images)
            to_show = images + folder_images
            if hasattr(self, "focused_image") and self.focused_image is not None:
                try:
                    to_show.remove(self.focused_image)
                except Exception:
                    pass
                to_show.insert(0, self.focused_image)
                self.focused_image = None
            self.parent.thumbs_manager.show(
                to_show, screen=self.get_screen(), folders=folders, type=thumbs_type
            )
            if pin:
                self.parent.thumbs_manager.pin()
            if thumbs_type:
                self.parent.update_indicator(auto_changed=False)

        except Exception:
            logger.exception(lambda: "Could not create thumbs window:")

    def on_add_configurable(self, source):
        if source.get_source_type() == Options.SourceType.WALLHAVEN:
            dialog = AddWallhavenDialog(self.parent)
        else:
            dialog = AddConfigurableDialog()
        dialog.set_source(source)
        self.show_dialog(dialog)

    def on_wallpaper_display_mode_changed(self, *args):
        modes = [
            m
            for m in self.parent.get_display_modes()
            if m.id == self.ui.wallpaper_display_mode.get_active_id()
        ]
        if modes:
            self.ui.wallpaper_mode_description.set_text(modes[0].description)
        else:
            self.ui.wallpaper_mode_description.set_text("")

    def show_dialog(self, dialog):
        self.dialog = dialog
        self.dialog.parent = self
        self.dialog.set_transient_for(self)
        response = self.dialog.run()
        if response != Gtk.ResponseType.OK:
            if self.dialog:
                self.dialog.destroy()
            self.dialog = None

    def on_add_dialog_okay(self, source_type, location, edited_row):
        if edited_row:
            edited_row[2] = location
        else:
            self.add_sources(source_type, [location])
        self.dialog = None

    def close(self):
        self.ui.error_favorites.set_label("")
        self.ui.error_fetched.set_label("")

        self.hide()
        self.parent.trigger_download()
        self.on_destroy()

    def on_save_clicked(self, widget):
        self.delayed_apply()
        self.close()

    def delayed_apply(self, widget=None, *arg):
        if not self.loading:
            self.delayed_apply_with_interval(0.1)

    def delayed_apply_slow(self, widget=None, *arg):
        if not self.loading:
            self.delayed_apply_with_interval(1)

    def delayed_apply_with_interval(self, interval):
        # pylint: disable=access-member-before-definition
        if not self.loading:
            if hasattr(self, "apply_timer") and self.apply_timer:
                self.apply_timer.cancel()
                self.apply_timer = None

            self.apply_timer = threading.Timer(interval, self.apply)
            self.apply_timer.start()

    def apply(self):
        try:
            logger.info(lambda: "Applying preferences")

            self.options = Options()
            self.options.read()

            self.options.change_enabled = self.ui.change_enabled.get_active()
            self.options.change_on_start = self.ui.change_on_start.get_active()
            self.options.change_interval = self.get_change_interval()
            self.options.internet_enabled = self.ui.internet_enabled.get_active()

            if os.access(self.fav_chooser.get_folder(), os.W_OK):
                self.options.favorites_folder = self.fav_chooser.get_folder()
            self.options.favorites_operations = self.favorites_operations

            self.options.sources = []
            for r in self.ui.sources.get_model():
                self.options.sources.append(self.model_row_to_source(r))
            for s in self.unsupported_sources:
                self.options.sources.append(s)

            self.options.wallpaper_auto_rotate = self.ui.wallpaper_auto_rotate.get_active()
            self.options.wallpaper_display_mode = self.ui.wallpaper_display_mode.get_active_id()

            if os.access(self.fetched_chooser.get_folder(), os.W_OK):
                self.options.fetched_folder = self.fetched_chooser.get_folder()
            self.options.clipboard_enabled = self.ui.clipboard_enabled.get_active()
            self.options.clipboard_use_whitelist = self.ui.clipboard_use_whitelist.get_active()
            buf = self.ui.clipboard_hosts.get_buffer()
            self.options.clipboard_hosts = Util.split(
                buf.get_text(buf.get_start_iter(), buf.get_end_iter(), False)
            )

            if self.ui.icon.get_active() == 0:
                self.options.icon = "Light"
            elif self.ui.icon.get_active() == 1:
                self.options.icon = "Dark"
            elif self.ui.icon.get_active() == 2:
                self.options.icon = "1"
            elif self.ui.icon.get_active() == 3:
                self.options.icon = "2"
            elif self.ui.icon.get_active() == 4:
                self.options.icon = "3"
            elif self.ui.icon.get_active() == 5:
                self.options.icon = "4"
            elif self.ui.icon.get_active() == 6:
                self.options.icon = "Current"
            elif self.ui.icon.get_active() == 8:
                self.options.icon = "None"
            elif self.ui.icon.get_active() == 7:
                file = self.ui.icon_chooser.get_filename()
                if file and os.access(file, os.R_OK):
                    self.options.icon = file
                else:
                    self.options.icon = "Light"

            if self.ui.favorites_operations.get_active() == 0:
                self.options.favorites_operations = [["/", "Copy"]]
            elif self.ui.favorites_operations.get_active() == 1:
                self.options.favorites_operations = [["/", "Move"]]
            elif self.ui.favorites_operations.get_active() == 2:
                self.options.favorites_operations = [["/", "Both"]]
            elif self.ui.favorites_operations.get_active() == 3:
                # will be set in the favops editor dialog
                pass

            self.options.copyto_enabled = self.ui.copyto_enabled.get_active()
            self.options.change_lock_screen = self.ui.change_lock_screen_box.get_active()
            copyto = os.path.normpath(self.copyto_chooser.get_folder())
            if copyto == os.path.normpath(self.parent.get_actual_copyto_folder("Default")):
                self.options.copyto_folder = "Default"
            else:
                self.options.copyto_folder = copyto

            self.options.desired_color_enabled = self.ui.desired_color_enabled.get_active()
            c = self.ui.desired_color.get_color()
            self.options.desired_color = (c.red // 256, c.green // 256, c.blue // 256)

            self.options.min_size_enabled = self.ui.min_size_enabled.get_active()
            try:
                self.options.min_size = int(self.ui.min_size.get_active_text())
            except Exception:
                pass

            self.options.use_landscape_enabled = self.ui.landscape_enabled.get_active()

            self.options.lightness_enabled = self.ui.lightness_enabled.get_active()
            self.options.lightness_mode = (
                Options.LightnessMode.DARK
                if self.ui.lightness.get_active() == 0
                else Options.LightnessMode.LIGHT
            )

            self.options.min_rating_enabled = self.ui.min_rating_enabled.get_active()
            try:
                self.options.min_rating = int(self.ui.min_rating.get_active_text())
            except Exception:
                pass

            self.options.name_regex_enabled = self.ui.name_regex_enabled.get_active()
            try:
                self.options.name_regex = self.ui.name_regex.get_text()
            except Exception:
                pass

            self.options.clock_enabled = self.ui.clock_enabled.get_active()
            self.options.clock_font = self.ui.clock_font.get_font_name()
            self.options.clock_date_font = self.ui.clock_date_font.get_font_name()

            self.options.quotes_enabled = self.ui.quotes_enabled.get_active()
            self.options.quotes_font = self.ui.quotes_font.get_font_name()
            c = self.ui.quotes_text_color.get_color()
            self.options.quotes_text_color = (c.red // 256, c.green // 256, c.blue // 256)
            c = self.ui.quotes_bg_color.get_color()
            self.options.quotes_bg_color = (c.red // 256, c.green // 256, c.blue // 256)
            self.options.quotes_bg_opacity = max(
                0, min(100, int(self.ui.quotes_bg_opacity.get_value()))
            )
            self.options.quotes_text_shadow = self.ui.quotes_text_shadow.get_active()
            self.options.quotes_tags = self.ui.quotes_tags.get_text()
            self.options.quotes_authors = self.ui.quotes_authors.get_text()
            self.options.quotes_change_enabled = self.ui.quotes_change_enabled.get_active()
            self.options.quotes_change_interval = self.get_quotes_change_interval()
            self.options.quotes_width = max(0, min(100, int(self.ui.quotes_width.get_value())))
            self.options.quotes_hpos = max(0, min(100, int(self.ui.quotes_hpos.get_value())))
            self.options.quotes_vpos = max(0, min(100, int(self.ui.quotes_vpos.get_value())))

            self.options.quotes_disabled_sources = [
                cb.get_label() for cb in self.quotes_sources_checkboxes if not cb.get_active()
            ]

            for f in self.options.filters:
                f[0] = self.filter_name_to_checkbox[f[1]].get_active()

            self.options.slideshow_sources_enabled = self.ui.slideshow_sources_enabled.get_active()
            self.options.slideshow_favorites_enabled = (
                self.ui.slideshow_favorites_enabled.get_active()
            )
            self.options.slideshow_downloads_enabled = (
                self.ui.slideshow_downloads_enabled.get_active()
            )
            self.options.slideshow_custom_enabled = self.ui.slideshow_custom_enabled.get_active()
            if os.access(self.slideshow_custom_chooser.get_folder(), os.R_OK):
                self.options.slideshow_custom_folder = self.slideshow_custom_chooser.get_folder()

            if self.ui.slideshow_sort_order.get_active() == 0:
                self.options.slideshow_sort_order = "Random"
            elif self.ui.slideshow_sort_order.get_active() == 1:
                self.options.slideshow_sort_order = "Name, asc"
            elif self.ui.slideshow_sort_order.get_active() == 2:
                self.options.slideshow_sort_order = "Name, desc"
            elif self.ui.slideshow_sort_order.get_active() == 3:
                self.options.slideshow_sort_order = "Date, asc"
            elif self.ui.slideshow_sort_order.get_active() == 4:
                self.options.slideshow_sort_order = "Date, desc"

            if self.ui.slideshow_monitor.get_active() == 0:
                self.options.slideshow_monitor = "All"
            else:
                self.options.slideshow_monitor = self.ui.slideshow_monitor.get_active()

            if self.ui.slideshow_mode.get_active() == 0:
                self.options.slideshow_mode = "Fullscreen"
            elif self.ui.slideshow_mode.get_active() == 1:
                self.options.slideshow_mode = "Desktop"
            elif self.ui.slideshow_mode.get_active() == 2:
                self.options.slideshow_mode = "Maximized"
            elif self.ui.slideshow_mode.get_active() == 3:
                self.options.slideshow_mode = "Window"

            self.options.slideshow_seconds = max(0.5, float(self.ui.slideshow_seconds.get_value()))
            self.options.slideshow_fade = max(0, min(1, float(self.ui.slideshow_fade.get_value())))
            self.options.slideshow_zoom = max(0, min(1, float(self.ui.slideshow_zoom.get_value())))
            self.options.slideshow_pan = max(0, min(0.2, float(self.ui.slideshow_pan.get_value())))

            self.options.write()

            if not self.parent.running:
                return

            self.parent.reload_config()

            self.update_autostart()
        except Exception:
            if self.parent.running:
                logger.exception(lambda: "Error while applying preferences")
                dialog = Gtk.MessageDialog(
                    self,
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.ERROR,
                    Gtk.ButtonsType.OK,
                    "An error occurred while saving preferences.\n"
                    "Please run from a terminal with the -v flag and try again.",
                )
                dialog.set_title("Oops")
                dialog.run()
                dialog.destroy()

    def update_autostart(self):
        file = get_autostart_file_path()

        if not self.ui.autostart.get_active():
            if os.path.exists(file):
                logger.info(lambda: "Removing autostart entry")
                Util.safe_unlink(file)
        else:
            if not os.path.exists(file):
                self.parent.create_autostart_entry()

    def on_change_enabled_toggled(self, widget=None):
        self.ui.change_interval_text.set_sensitive(self.ui.change_enabled.get_active())
        self.ui.change_interval_time_unit.set_sensitive(self.ui.change_enabled.get_active())

    def on_quotes_change_enabled_toggled(self, widget=None):
        self.ui.quotes_change_interval_text.set_sensitive(
            self.ui.quotes_change_enabled.get_active()
        )
        self.ui.quotes_change_interval_time_unit.set_sensitive(
            self.ui.quotes_change_enabled.get_active()
        )

    def on_desired_color_enabled_toggled(self, widget=None):
        self.ui.desired_color.set_sensitive(self.ui.desired_color_enabled.get_active())

    def on_min_size_enabled_toggled(self, widget=None):
        self.ui.min_size.set_sensitive(self.ui.min_size_enabled.get_active())
        self.ui.min_size_label.set_sensitive(self.ui.min_size_enabled.get_active())

    def on_min_rating_enabled_toggled(self, widget=None):
        self.ui.min_rating.set_sensitive(self.ui.min_rating_enabled.get_active())

    def on_name_regex_enabled_toggled(self, widget=None):
        self.ui.name_regex.set_sensitive(self.ui.name_regex_enabled.get_active())

    def on_lightness_enabled_toggled(self, widget=None):
        self.ui.lightness.set_sensitive(self.ui.lightness_enabled.get_active())

    def on_destroy(self, widget=None):
        if hasattr(self, "dialog") and self.dialog:
            try:
                self.dialog.destroy()
            except Exception:
                pass
        for chooser in (self.fav_chooser, self.fetched_chooser):
            try:
                chooser.destroy()
            except Exception:
                pass
        self.parent.thumbs_manager.hide(force=False)

    def on_favorites_changed(self, widget=None):
        self.delayed_apply()
        if not os.access(self.fav_chooser.get_folder(), os.W_OK):
            self.ui.error_favorites.set_label(_("No write permissions"))
        else:
            self.ui.error_favorites.set_label("")

    def on_fetched_changed(self, widget=None):
        self.delayed_apply()
        if not os.access(self.fetched_chooser.get_folder(), os.W_OK):
            self.ui.error_fetched.set_label(_("No write permissions"))
        else:
            self.ui.error_fetched.set_label("")

    def update_clipboard_state(self, widget=None):
        self.ui.clipboard_use_whitelist.set_sensitive(self.ui.clipboard_enabled.get_active())
        # keep the hosts list always enabled - user can wish to add hosts even when monitoring is not enabled - if undesired, uncomment below:
        # self.ui.clipboard_hosts.set_sensitive(self.ui.clipboard_enabled.get_active() and self.ui.clipboard_use_whitelist.get_active())

    def on_edit_favorites_operations_clicked(self, widget=None):
        self.dialog = EditFavoriteOperationsDialog()
        self.dialog.set_transient_for(self)
        buf = self.dialog.ui.textbuffer
        buf.set_text("\n".join(":".join(x) for x in self.favorites_operations))
        if self.dialog.run() == Gtk.ResponseType.OK:
            text = buf.get_text(buf.get_start_iter(), buf.get_end_iter(), False)
            self.favorites_operations = list([x.strip().split(":") for x in text.split("\n") if x])
            self.delayed_apply()
        self.dialog.destroy()
        self.dialog = None

    def on_icon_changed(self, widget=None):
        self.ui.icon_chooser.set_visible(self.ui.icon.get_active() == 7)

    def on_favorites_operations_changed(self, widget=None):
        self.ui.edit_favorites_operations.set_visible(
            self.ui.favorites_operations.get_active() == 3
        )

    def on_copyto_enabled_toggled(self, widget=None):
        self.copyto_chooser.set_sensitive(self.ui.copyto_enabled.get_active())
        self.ui.copyto_use_default.set_sensitive(self.ui.copyto_enabled.get_active())
        self.on_copyto_changed()

    def on_copyto_changed(self):
        self.ui.copyto_faq_link.set_sensitive(self.ui.copyto_enabled.get_active())
        if self.ui.copyto_enabled.get_active() and self.copyto_chooser.get_folder():
            folder = self.copyto_chooser.get_folder()
            self.ui.copyto_use_default.set_sensitive(
                folder != self.parent.get_actual_copyto_folder("Default")
            )
            under_encrypted = Util.is_home_encrypted() and folder.startswith(
                os.path.expanduser("~") + "/"
            )
            self.ui.copyto_encrypted_note.set_visible(under_encrypted)
            can_write = os.access(self.parent.get_actual_copyto_folder(folder), os.W_OK)
            can_read = os.stat(folder).st_mode | stat.S_IROTH
            self.ui.copyto_faq_link.set_visible(can_write and can_read and not under_encrypted)
            self.ui.copyto_permissions_box.set_visible(not can_write or not can_read)
            self.ui.copyto_write_permissions_warning.set_visible(not can_write)
            self.ui.copyto_read_permissions_warning.set_visible(not can_read)
        else:
            self.ui.copyto_faq_link.set_visible(True)
            self.ui.copyto_encrypted_note.set_visible(False)
            self.ui.copyto_permissions_box.set_visible(False)
        self.delayed_apply()

    def on_copyto_use_default_clicked(self, widget=None):
        self.copyto_chooser.set_folder(self.parent.get_actual_copyto_folder("Default"))
        self.on_copyto_changed()

    def on_copyto_fix_permissions_clicked(self, widget=None):
        folder = self.copyto_chooser.get_folder()
        can_write = os.access(self.parent.get_actual_copyto_folder(folder), os.W_OK)
        can_read = os.stat(folder).st_mode | stat.S_IROTH
        mode = "a+"
        if not can_read:
            mode += "r"
        if not can_write:
            mode += "w"
        try:
            Util.superuser_exec("chmod", mode, folder)
        except Exception:
            logger.exception(lambda: "Could not adjust copyto folder permissions")
            self.parent.show_notification(
                _("Could not adjust permissions"),
                _('You may try manually running this command:\nsudo chmod %s "%s"')
                % (mode, folder),
            )
        self.on_copyto_changed()

    def on_btn_slideshow_reset_clicked(self, widget=None):
        self.ui.slideshow_seconds.set_value(6)
        self.ui.slideshow_fade.set_value(0.4)
        self.ui.slideshow_zoom.set_value(0.2)
        self.ui.slideshow_pan.set_value(0.05)

    def on_btn_slideshow_start_clicked(self, widget=None):
        self.apply()
        self.parent.on_start_slideshow()
